# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).

from . import ioHubKeyboardDevice
from .. import Computer, Device
from ...constants import EventConstants
from ...errors import printExceptionDetailsToStdErr

getTime = Computer.getTime


class Keyboard(ioHubKeyboardDevice):
    event_id_index = None

    def __init__(self, *args, **kwargs):
        ioHubKeyboardDevice.__init__(self, *args, **kwargs['dconfig'])

        if self.event_id_index is None:
            from . import KeyboardInputEvent
            Keyboard.auto_repeated_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'auto_repeated')
            Keyboard.key_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'key')
            Keyboard.key_id_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'key_id')
            Keyboard.event_type_index = KeyboardInputEvent.EVENT_TYPE_ID_INDEX
            Keyboard.event_modifiers_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'modifiers')
            Keyboard.win_id_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'window_id')
            Keyboard.event_id_index = KeyboardInputEvent.EVENT_ID_INDEX

    def _nativeEventCallback(self, event):
        try:
            self._last_callback_time = getTime()
            if self.isReportingEvents():
                event_array = event[0]

                # Check if key event window id is in list of psychopy
                # windows and what report_system_wide_events value is
                report_system_wide_events = self.getConfiguration().get(
                    'report_system_wide_events', True)
                if report_system_wide_events is False:
                    pyglet_window_hnds = self._iohub_server._pyglet_window_hnds
                    if len(pyglet_window_hnds) > 0 and event_array[
                            self.win_id_index] not in pyglet_window_hnds:
                        return True

                is_pressed = event_array[
                    self.event_type_index] == EventConstants.KEYBOARD_PRESS

                if is_pressed and self._report_auto_repeats is False:
                    # AUto repeat value provided by pyXHook code
                    auto_repeat_count = event_array[self.auto_repeated_index]
                    if auto_repeat_count > 0:
                        return True

                # set event id for event since it has passed all filters
                event_array[self.event_id_index] = Device._getNextEventID()
                ioHubKeyboardDevice._modifier_value = event_array[
                    self.event_modifiers_index]

                self._updateKeyboardEventState(event_array, is_pressed)

                self._addNativeEventToBuffer(event_array)
        except Exception:
            printExceptionDetailsToStdErr()
        return 1

    def _getIOHubEventObject(self, native_event_data):
        return native_event_data



#!/usr/bin/env python
#  -*- coding: utf-8 -*-

# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
from __future__ import division, absolute_import, print_function

import sys

from .errors import print2err, printExceptionDetailsToStdErr
from .util import module_directory, fix_encoding

fix_encoding.fix_encoding()

if sys.platform == 'darwin':
    import objc  # pylint: disable=import-error

EXP_SCRIPT_DIRECTORY = ''

def _localFunc():
    return None
IOHUB_DIRECTORY = module_directory(_localFunc)

_ispkg = True
_pkgroot = 'iohub'
if IOHUB_DIRECTORY.find('psychopy') >= 0:
    _ispkg = False
    _pkgroot = 'psychopy.iohub'

_DATA_STORE_AVAILABLE = False
try:
    import tables # pylint: disable=wrong-import-position, wrong-import-order
    _DATA_STORE_AVAILABLE = True
except ImportError:
    print2err('WARNING: pytables package not found. ',
              'ioHub functionality will be disabled.')
except Exception: # pylint: disable=broad-except
    printExceptionDetailsToStdErr()

from .client.expruntime import ioHubExperimentRuntime

lazyImports = """
from {pkgroot}.client.connect import launchHubServer
from {pkgroot}.devices.computer import Computer
""".format(pkgroot=_pkgroot)

try:
    from psychopy.contrib.lazy_import import lazy_import
    lazy_import(globals(), lazyImports)
except Exception:
    exec(lazyImports)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file was generated by generateHints.py.
# Following strings are used to localize hints in Preference Dialog of
# the PsychoPy application.
# Rebuild this file if comments in *.spec files are modified.

from __future__ import absolute_import, print_function
from psychopy.localization import _translate

# baseNoArch.spec,[general],line25
_translate("which system to use as a backend for drawing")

# baseNoArch.spec,[general],line27
_translate("the default units for windows and visual stimuli")

# baseNoArch.spec,[general],line29
_translate("full screen is best for accurate timing")

# baseNoArch.spec,[general],line31
_translate("enable subjects to use the mouse and GUIs during experiments")

# baseNoArch.spec,[general],line33
_translate("'version' is for internal usage, not for the user")

# baseNoArch.spec,[general],line35
_translate("Add paths here to your custom Python modules")

# baseNoArch.spec,[general],line37
_translate("path to flac (lossless audio compression) on this operating system")

# baseNoArch.spec,[general],line39
_translate("Shutdown keys, following the pyglet naming scheme.")

# baseNoArch.spec,[general],line41
_translate("Modifier keys for shutdown keys")

# baseNoArch.spec,[general],line43
_translate("What to do if gamma-correction not possible")

# baseNoArch.spec,[general],line45
_translate("Add plugin names here to load when a PsychoPy session starts.")

# baseNoArch.spec,[general],line48
_translate("display tips when starting PsychoPy")

# baseNoArch.spec,[app],line50
_translate("what windows to display when PsychoPy starts")

# baseNoArch.spec,[app],line52
_translate("reset preferences to defaults on next restart of PsychoPy")

# baseNoArch.spec,[app],line54
_translate("save any unsaved preferences before closing the window")

# baseNoArch.spec,[app],line56
_translate("enable features for debugging PsychoPy itself, including unit-tests")

# baseNoArch.spec,[app],line58
_translate("language to use in menus etc; not all translations are available. Select a value, then restart the app.")

# baseNoArch.spec,[app],line60
_translate(
    "Show an error dialog when PsychoPy encounters an unhandled internal error.")

# baseNoArch.spec,[app],line62
_translate("Theme")

# baseNoArch.spec,[app],line64
_translate(
    "open Coder files as read-only (allows running without accidental changes)")

# baseNoArch.spec,[app],line67
_translate("a list of font names; the first one found on the system will be used")

# baseNoArch.spec,[coder],line69
_translate("Font size (in pts) takes an integer between 6 and 24")

# baseNoArch.spec,[coder],line71
_translate("Spacing between lines")

# baseNoArch.spec,[coder],line75
_translate("Long line edge guide, specify zero to disable")

# baseNoArch.spec,[coder],line79
_translate("Set the source assistant panel to be visible by default")

# baseNoArch.spec,[coder],line81
_translate("Set the output/shell to be visible by default")

# baseNoArch.spec,[coder],line83
_translate(
    "Show code completion suggestion and calltips automatically when typing.")

# baseNoArch.spec,[coder],line85
_translate("reload previously opened files after start")

# baseNoArch.spec,[coder],line87
_translate("for coder shell window, which shell to use")

# baseNoArch.spec,[coder],line89
_translate("whether to automatically reload a previously open experiment")

# baseNoArch.spec,[coder],line91
_translate("Default to when writing code components")

# baseNoArch.spec,[coder],line94
_translate(
    "if False will create scripts with an 'easier' but more cluttered namespace")

# baseNoArch.spec,[builder],line96
_translate("folder names for custom components; expects a comma-separated list")

# baseNoArch.spec,[builder],line98
_translate("a list of components to hide (eg, because you never use them)")

# baseNoArch.spec,[builder],line100
_translate("where the Builder demos are located on this computer (after unpacking)")

# baseNoArch.spec,[builder],line102
_translate(
    "name of the folder where subject data should be saved (relative to the script)")

# baseNoArch.spec,[builder],line104
_translate("Panels arrangement: Should Flow be on the top or bottom, and should Components be on the left or right?")

# baseNoArch.spec,[builder],line106
_translate("Display text in a floating window that describes the experiment")

# baseNoArch.spec,[builder],line108
_translate("choice of audio library")

# baseNoArch.spec,[builder],line110
_translate(
    "latency mode for PsychToolbox audio (3 is good for most applications. See")

# baseNoArch.spec,[builder],line112
_translate("audio driver to use")

# baseNoArch.spec,[builder],line114
_translate("audio device to use (if audioLib allows control)")

# baseNoArch.spec,[hardware],line118
_translate("a list of parallel ports")

# baseNoArch.spec,[hardware],line120
_translate("The name of the Qmix pump configuration to use")

# baseNoArch.spec,[hardware],line122
_translate("the http proxy for usage stats and auto-updating; format is host: port")

# baseNoArch.spec,[hardware],line124
_translate(
    "override the above proxy settings with values found in the environment (if possible)")

# baseNoArch.spec,[hardware],line126
_translate("allow PsychoPy to send anonymous usage stats; please allow if possible, it helps PsychoPy's development")

# baseNoArch.spec,[hardware],line128
_translate("allow PsychoPy to check for new features and bug fixes")

# baseNoArch.spec,[hardware],line131
_translate("max time to wait for a connection response")

# baseNoArch.spec,[connections],line133
_translate("open an existing file")

# baseNoArch.spec,[connections],line135
_translate("start a new experiment or script")

# baseNoArch.spec,[connections],line137
_translate("save a Builder or Coder file")

# baseNoArch.spec,[connections],line139
_translate("save a Builder or Coder file under a new name")

# baseNoArch.spec,[connections],line141
_translate("Coder: print the file")

# baseNoArch.spec,[connections],line144
_translate("close the Builder or Coder window")

# baseNoArch.spec,[keyBindings],line146
_translate("end the application (PsychoPy)")

# baseNoArch.spec,[keyBindings],line148
_translate("open the preferences dialog")

# baseNoArch.spec,[keyBindings],line150
_translate("export Builder experiment to HTML")

# baseNoArch.spec,[keyBindings],line152
_translate("Coder: cut")

# baseNoArch.spec,[keyBindings],line154
_translate("Coder: copy")

# baseNoArch.spec,[keyBindings],line156
_translate("Coder: paste")

# baseNoArch.spec,[keyBindings],line158
_translate("Coder: duplicate")

# baseNoArch.spec,[keyBindings],line160
_translate("Coder: indent code by one level (4 spaces)")

# baseNoArch.spec,[keyBindings],line162
_translate("Coder: reduce indentation by one level (4 spaces)")

# baseNoArch.spec,[keyBindings],line165
_translate("Coder: indent to fit python syntax")

# baseNoArch.spec,[keyBindings],line167
_translate("Coder: find")

# baseNoArch.spec,[keyBindings],line169
_translate("Coder: find again")

# baseNoArch.spec,[keyBindings],line171
_translate("Coder: undo")

# baseNoArch.spec,[keyBindings],line173
_translate("Coder: redo")

# baseNoArch.spec,[keyBindings],line175
_translate("Coder: add a # to the start of the line(s)")

# baseNoArch.spec,[keyBindings],line177
_translate("Coder: remove # from start of line(s)")

# baseNoArch.spec,[keyBindings],line179
_translate("Coder: add or remove # from start of line(s)")

# baseNoArch.spec,[keyBindings],line181
_translate("Coder: fold this block of code")

# baseNoArch.spec,[keyBindings],line183
_translate("Coder: increase font size this block of code")

# baseNoArch.spec,[keyBindings],line185
_translate("Coder: decrease font size this block of code")

# baseNoArch.spec,[keyBindings],line187
_translate("Coder: check for basic syntax errors")

# baseNoArch.spec,[keyBindings],line189
_translate(
    "convert a Builder .psyexp script into a python script and open it in the Coder")

# baseNoArch.spec,[keyBindings],line191
_translate("launch a script, Builder or Coder, or run unit-tests")

# baseNoArch.spec,[keyBindings],line193
_translate("attempt to interrupt and halt a running script")

# baseNoArch.spec,[keyBindings],line195
_translate("Coder: show / hide white-space dots")

# baseNoArch.spec,[keyBindings],line197
_translate("Coder: show / hide end of line characters")

# baseNoArch.spec,[keyBindings],line200
_translate("Coder: show / hide indentation level lines")

# baseNoArch.spec,[keyBindings],line202
_translate("Builder: create a new routine")

# baseNoArch.spec,[keyBindings],line204
_translate("Builder: copy an existing routine")

# baseNoArch.spec,[keyBindings],line208
_translate("Builder: paste the copied routine")

# baseNoArch.spec,[keyBindings],line211
_translate("Builder: paste the copied component")

# baseNoArch.spec,[keyBindings],line213
_translate("Coder: show / hide the output panel")

# baseNoArch.spec,[keyBindings],line215
_translate("Builder: rename an existing routine")

# baseNoArch.spec,[keyBindings],line218
_translate("switch between windows")

# baseNoArch.spec,[keyBindings],line220
_translate("increase display size in Flow")

# baseNoArch.spec,[keyBindings],line222
_translate("decrease display size in Flow")

# baseNoArch.spec,[keyBindings],line224
_translate("increase display size of Routines")

# baseNoArch.spec,[keyBindings],line226
_translate("decrease display size of Routines")

# baseNoArch.spec,[keyBindings],line228
_translate("show or hide the readme (info) for this experiment if possible")

# baseNoArch.spec,[keyBindings],line230
_translate("Projects: Log in to pavlovia")

# baseNoArch.spec,[keyBindings],line232
_translate("Projects: Log in to OSF")

# baseNoArch.spec,[keyBindings],line234
_translate("Projects: Sync project")

# baseNoArch.spec,[keyBindings],line236
_translate("Projects: Find projects")

# baseNoArch.spec,[keyBindings],line238
_translate("Projects: Open project")

# baseNoArch.spec,[keyBindings],line240
_translate("Projects: Create new project")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Extensible set of components for the PsychoPy Builder view
"""

from __future__ import absolute_import, print_function

from builtins import str
from past.builtins import basestring
import os
import glob
import copy
import shutil
from os.path import join, dirname, abspath, split
from importlib import import_module  # helps python 2.7 -> 3.x migration
from ._base import BaseVisualComponent, BaseComponent
from ..params import Param
from psychopy.localization import _translate
from psychopy.experiment import py2js

excludeComponents = ['BaseComponent', 'BaseVisualComponent',  # templates only
                     'EyetrackerComponent']  # this one isn't ready yet

pluginComponents = {}  # components registered by loaded plugins

# try to remove old pyc files in case they're detected as components
pycFiles = glob.glob(join(split(__file__)[0], "*.pyc"))
for filename in pycFiles:
    # check for matching py file
    if not os.path.isfile(filename[:-2]):
        try:
            os.remove(filename)
        except:
            pass  # may not have sufficient privs


def getAllCategories(folderList=()):
    allComps = getAllComponents(folderList)
    allCats = ['Stimuli', 'Responses', 'Custom']
    for name, thisComp in list(allComps.items()):
        for thisCat in thisComp.categories:
            if thisCat not in allCats:
                allCats.append(thisCat)
    return allCats


def getAllComponents(folderList=(), fetchIcons=True):
    """Get a dictionary of all available components, from the builtins as well
    as all folders in the folderlist.
    User-defined components will override built-ins with the same name.
    """
    if isinstance(folderList, basestring):
        raise TypeError('folderList should be iterable, not a string')
    components = getComponents(fetchIcons=fetchIcons)  # get the built-ins
    for folder in folderList:
        userComps = getComponents(folder)
        for thisKey in userComps:
            components[thisKey] = userComps[thisKey]

    # add components registered by plugins that have been loaded
    components.update(pluginComponents)

    return components


def getComponents(folder=None, fetchIcons=True):
    """Get a dictionary of available components for the Builder experiments.

    If folder is None then the built-in components will be imported and
    returned, otherwise the components found in the folder provided will be.

    Changed v1.84.00:
    The Builder preference "components folders" should be of the form:
    `/.../.../compts`. This is unchanged from previously, and allows for
    backwards compatibility.

    New as of v1.84: A slightly different directory structure is needed. An
    existing directory will be automatically upgraded if the previous one
    was not empty. However, files starting with '_' will be skipped, as will any
    directories. You will need to manually move those into the new
    directory (from the old .../compts/ into the new .../compts/compts/).

    As of v1.84, a components path needs directory structure `/.../.../compts/compts`.
    That is, the path should end with the name repeated. (It does not need to
    be 'compts' literally.)
    The .py and .png files for a component should all go in this directory.
    (Previously, files were in `/.../.../compts`.) In addition, the directory
    must contain a python init file, ` /.../.../compts/compts/__init__.py`,
    to allow it to be treated as a module in python so that the components
    can be imported. For this reason, the file name for a component
    cannot begin with a number; it must be a legal python name.

    The code within the component.py file itself must use absolute paths for
    importing from psychopy:
       `from psychopy.experiment.components import BaseComponent, Param`
    """
    if folder is None:
        pth = folder = dirname(__file__)
        pkg = 'psychopy.experiment.components'
    else:
        # default shared location is often not actually a folder
        if not os.path.isdir(folder):
            return {}
        pth = folder = folder.rstrip(os.sep)
        pkg = os.path.basename(folder)
        if not folder.endswith(join(pkg, pkg)):
            folder = os.path.join(folder, pkg)

        # update the old style directory (v1.83.03) to the new style
        # try to retain backwards compatibility: copy files, not move them
        # ideally hard link them, but permissions fail on windows
        if not os.path.isdir(folder):
            files = [f for f in glob.glob(join(pth, '*'))
                     if not os.path.isdir(f) and
                     not f[0] in '_0123456789']
            if files:
                os.mkdir(folder)
                with open(join(folder, '__init__.py'), 'a') as fileh:
                    fileh.write('')
                for f in files:
                    if f.startswith('_'):
                        continue
                    shutil.copy(f, folder)
    if not pth in os.sys.path:
        os.sys.path.insert(0, pth)

    components = {}

    # go through components in directory
    cfiles = glob.glob(os.path.join(folder, '*.py'))  # old-style: just comp.py
    # new-style: directories w/ __init__.py
    dfiles = [d for d in os.listdir(folder)
              if os.path.isdir(os.path.join(folder, d))]
    for cmpfile in cfiles + dfiles:
        cmpfile = os.path.split(cmpfile)[1]
        if cmpfile[0] in '_0123456789':  # __init__.py, _base.py, leading digit
            continue
        # can't use imp - breaks py2app:
        # module = imp.load_source(file[:-3], fullPath)
        # v1.83.00 used exec(implicit-relative), no go for python3:
        # exec('import %s as module' % file[:-3])

        # importlib.import_module eases 2.7 -> 3.x migration
        if cmpfile.endswith('.py'):
            explicit_rel_path = pkg + '.' + cmpfile[:-3]
        else:
            explicit_rel_path = pkg + '.' + cmpfile
        try:
            module = import_module(explicit_rel_path, package=pkg)
        except ImportError:
            continue  # not a valid module (no __init__.py?)
        # check for orphaned pyc files (__file__ is not a .py file)
        if hasattr(module, '__file__'):
            if not module.__file__:
                # with Py3, orphans have a __pycharm__ folder but no file
                continue
            elif module.__file__.endswith('.pyc'):
                # with Py2, orphans have a xxxxx.pyc file
                if not os.path.isfile(module.__file__[:-1]):
                    continue  # looks like an orphaned pyc file
        # give a default category
        if not hasattr(module, 'categories'):
            module.categories = ['Custom']
        # check if module contains a component
        for attrib in dir(module):
            name = None
            # fetch the attribs that end with 'Component'
            if (attrib.endswith('omponent') and
                    attrib not in excludeComponents):
                name = attrib
                components[attrib] = getattr(module, attrib)

                # skip if this class was imported, not defined here
                if module.__name__ != components[attrib].__module__:
                    continue  # class was defined in different module

                if hasattr(module, 'tooltip'):
                    tooltips[name] = module.tooltip
                if hasattr(module, 'iconFile'):
                    iconFiles[name] = module.iconFile
                # assign the module categories to the Component
                if not hasattr(components[attrib], 'categories'):
                    components[attrib].categories = ['Custom']
    return components



def getInitVals(params, target="PsychoPy"):
    """Works out a suitable initial value for a parameter (e.g. to go into the
    __init__ of a stimulus object, avoiding using a variable name if possible
    """
    inits = copy.deepcopy(params)
    for name in params:
        if target == "PsychoJS":
            # convert (0,0.5) to [0,0.5] but don't convert "rand()" to "rand[]" and don't convert text
            valStr = str(inits[name].val).strip()
            if valStr.startswith("(") and valStr.endswith(")") and name != 'text':
                inits[name].val = py2js.expression2js(inits[name].val)
            # filenames (e.g. for image) need to be loaded from resources
            if name in ["sound"]:
                val = str(inits[name].val)
                if val not in [None, 'None', 'none', '']:
                    inits[name].val = ("psychoJS.resourceManager.getResource({})"
                                       .format(inits[name]))
                    inits[name].valType = 'code'

        if not hasattr(inits[name], 'updates'):  # might be settings parameter instead
            continue

        # value should be None (as code)
        elif inits[name].val in [None, 'None', 'none', '']:
            if name in ['text']:
                inits[name].val = None
                inits[name].valType = 'extendedStr'
            else:
                inits[name].val = 'None'
                inits[name].valType = 'code'

        # is constant so don't touch the parameter value
        elif inits[name].updates in ['constant', None, 'None']:
            continue  # things that are constant don't need handling

        # is changing so work out a reasonable default
        elif name in ['pos', 'fieldPos']:
            inits[name].val = '[0,0]'
            inits[name].valType = 'code'
        elif name in ['color', 'foreColor', 'borderColor', 'lineColor', 'fillColor']:
            inits[name].val = 'white'
            inits[name].valType = 'str'
        elif name in ['ori', 'sf', 'size', 'height', 'letterHeight', 'lineWidth',
                      'phase', 'opacity',
                      'volume',  # sounds
                      'coherence', 'nDots', 'fieldSize', 'dotSize', 'dotLife',
                      'dir', 'speed',
                      'contrast', 'moddepth', 'envori', 'envphase', 'envsf',
                      'noiseClip', 'noiseBWO', 'noiseFilterUpper', 'noiseFilterLower',
                      'noiseBaseSf', 'noiseBW', 'noiseElementSize', 'noiseFilterOrder',
                      'noiseFractalPower']:
            inits[name].val = "1.0"
            inits[name].valType = 'code'
        elif name in ['image', 'mask', 'envelope', 'carrier']:
            inits[name].val = "sin"
            inits[name].valType = 'str'
        elif name == 'texture resolution':
            inits[name].val = "128"
            inits[name].valType = 'code'
        elif name == 'colorSpace':
            inits[name].val = "rgb"
            inits[name].valType = 'str'
        elif name == 'font':
            inits[name].val = "Arial"
            inits[name].valType = 'str'
        elif name == 'units':
            inits[name].val = "norm"
            inits[name].valType = 'str'
        elif name == 'text':
            inits[name].val = ""
            inits[name].valType = 'str'
        elif name == 'flip':
            inits[name].val = ""
            inits[name].valType = 'str'
        elif name == 'sound':
            inits[name].val = "A"
            inits[name].valType = 'str'
        elif name == 'blendmode':
            inits[name].val = "avg"
            inits[name].valType = 'str'
        elif name == 'beat':
            inits[name].val = "False"
            inits[name].valType = 'str'
        elif name == 'noiseImage':
            inits[name].val = "None"
            inits[name].valType = 'str'
        elif name == 'noiseType':
            inits[name].val = 'Binary'
            inits[name].valType = 'str'
        elif name == 'marker_label':
            inits[name].val = 'Label'
            inits[name].valType = 'str'
        elif name == 'marker_value':
            inits[name].val = 'Value'
            inits[name].valType = 'str'
        elif name == 'buttonRequired':
            inits[name].val = "True"
            inits[name].valType = 'code'
        else:
            print("I don't know the appropriate default value for a '%s' "
                  "parameter. Please email the mailing list about this error" %
                  name)

    return inits

tooltips = {}
iconFiles = {}



import pytest

__author__ = 'Sol'

import psutil, sys
from psychopy.iohub import launchHubServer, Computer

getTime = Computer.getTime

from psychopy.tests.utils import skip_under_travis

@skip_under_travis
def startHubProcess():
    io = launchHubServer()
    assert io != None

    io_proc = Computer.getIoHubProcess()
    io_proc_pid = io_proc.pid
    assert io_proc != None and io_proc_pid > 0

    return io

@skip_under_travis
def stopHubProcess():
    from psychopy.iohub.client import ioHubConnection

    io = ioHubConnection.getActiveConnection()
    assert io != None

    io_proc = Computer.getIoHubProcess()
    io_proc_pid = io_proc.pid
    assert io_proc != None and psutil.pid_exists(io_proc_pid)

    # Stop iohub server, ending process.
    io.quit()

    # Enure iohub proc has terminated.
    assert not psutil.pid_exists(io_proc_pid)

    assert ioHubConnection.getActiveConnection() is None

skip_not_completed = pytest.mark.skipif("True",
                                       reason="Cannot be tested until the test is completed.")

skip_under_windoz = pytest.mark.skipif("sys.platform == 'win32'",
                                       reason="Cannot be tested under Windoz.")

skip_under_linux = pytest.mark.skipif("sys.platform.startswith('linux')",
                                       reason="Cannot be tested under Linux.")

skip_under_osx = pytest.mark.skipif("sys.platform == 'darwin'",
                                       reason="Cannot be tested under macOS.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Creates a button"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, print_function

import builtins

from psychopy import event
from psychopy.visual import TextBox2

__author__ = 'Anthony Haffey & Todd Parsons'

class ButtonStim(TextBox2):
    """A class for putting a button into your experiment. A button is essentially a TextBox with a Mouse component contained within it, making it easy to check whether it has been clicked on.

    """

    def __init__(self, win, text, font='Arvo',
                 pos=(0, 0), size=None, padding=None, anchor='center', units=None,
                 color='white', fillColor='darkgrey', borderColor=None, borderWidth=0, colorSpace='rgb', opacity=None,
                 letterHeight=None, bold=True, italic=False,
                 name=""
                 ):
        # Initialise TextBox
        TextBox2.__init__(self, win, text, font, name=name,
                                 pos=pos, size=size, padding=padding, anchor=anchor, units=units,
                                 color=color, fillColor=fillColor, borderColor=borderColor, borderWidth=borderWidth, colorSpace=colorSpace, opacity=opacity,
                                 letterHeight=letterHeight, bold=bold, italic=italic,
                                 alignment='center', editable=False)
        self.listener = event.Mouse(win=win)
        self.wasClicked = False # Attribute to save whether button was previously clicked
        # Arrays to store times of clicks on and off
        self.timesOn = []
        self.timesOff = []
    @property
    def numClicks(self):
        """How many times has this button been clicked on?"""
        return len(self.timesOn)

    @property
    def isClicked(self):
        """Is this button currently being clicked on?"""
        return bool(self.listener.isPressedIn(self))


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 28 11:20:49 2017

@author: mrbki
"""
from os import path
from psychopy.experiment.components import BaseComponent, getInitVals
from psychopy.localization import _translate, _localized as __localized
_localized = __localized.copy()

CORTEX_OBJ = 'cortex_obj'

thisFolder = path.abspath(path.dirname(__file__))
iconFile = path.join(thisFolder, 'emotiv_record.png')
tooltip = _translate('Initialize EMOTIV hardware connection')


class EmotivRecordingComponent(BaseComponent):  # or (VisualComponent)
    def __init__(self, exp, parentName, name='cortex_rec'):
        super(EmotivRecordingComponent, self).__init__(
            exp, parentName, name=name,
            startType='time (s)', startVal=0,
            stopType='duration (s)', stopVal=1,
            startEstim='', durationEstim='',
            saveStartStop=False
        )
        self.exp.requireImport(importName='emotiv',
                               importFrom='psychopy.hardware')
        self.type = 'EmotivRecording'

    def writeInitCode(self, buff):
        inits = getInitVals(self.params, 'PsychoPy')
        code = ('{} = visual.BaseVisualStim('.format(inits['name']) +
                'win=win, name="{}")\n'.format(inits['name'])
                )
        buff.writeIndentedLines(code)
        code = ("{} = emotiv.Cortex(subject=expInfo['participant'])\n"
                .format(CORTEX_OBJ))
        buff.writeIndentedLines(code)

    def writeFrameCode(self, buff):
        pass

    def writeExperimentEndCode(self, buff):
        code = (
                "core.wait(1) # Wait for EEG data to be packaged\n" +
                "{}.close_session()\n".format(CORTEX_OBJ)
        )
        buff.writeIndentedLines(code)



from builtins import object
from pathlib import Path

from psychopy import visual, event
from psychopy.visual import Window
from psychopy.visual import TextBox2
from psychopy.visual.textbox2.fontmanager import FontManager
import pytest
from psychopy.tests import utils

# cd psychopy/psychopy
# py.test -k textbox --cov-report term-missing --cov visual/textbox

@pytest.mark.textbox
class Test_textbox(object):
    def setup_class(self):
        self.win = Window([128,128], pos=[50,50], allowGUI=False, autoLog=False)

    def teardown_class(self):
        self.win.close()

    def test_glyph_rendering(self):
        textbox = TextBox2(self.win, "", "Arial", pos=(0,0), size=(1,1), letterHeight=0.1, units='height')
        # Add all Noto Sans fonts to cover widest possible base of handles characters
        for font in ["Noto Sans", "Noto Sans HK", "Noto Sans JP", "Noto Sans KR", "Noto Sans SC", "Noto Sans TC", "Niramit", "Indie Flower"]:
            textbox.fontMGR.addGoogleFont(font)
        # Some exemplar text to test basic TextBox rendering
        exemplars = [
            # An English pangram
            {"text": "A PsychoPy zealot knows a smidge of wx, but JavaScript is the question.",
             "font": "Noto Sans",
             "screenshot": "textbox_exemplar_1.png"},
            # The same pangram in IPA
            {"text": "ə saɪkəʊpaɪ zɛlət nəʊz ə smidge ɒv wx, bʌt ˈʤɑːvəskrɪpt ɪz ðə ˈkwɛsʧən",
                "font": "Noto Sans",
                "screenshot": "textbox_exemplar_2.png"},
            # The same pangram in Hangul
            {"text": "아 프시초피 제알롣 크노W스 아 s믿게 오f wx, 붇 자v앗c립t 잇 테 q왯디온",
             "font": "Noto Sans KR",
             "screenshot": "textbox_exemplar_3.png"},
            # A noticeably non-standard font
            {"text": "A PsychoPy zealot knows a smidge of wx, but JavaScript is the question.",
             "font": "Indie Flower",
             "screenshot": "textbox_exemplar_4.png",
            }
        ]
        # Some text which is likely to cause problems if something isn't working
        tykes = [
            # Text which doesn't render properly on Mac (Issue #3203)
            {"text": "कोशिकायें",
             "font": "Noto Sans",
             "screenshot": "textbox_tyke_1.png"},
            # Thai text which old Text component couldn't handle due to Pyglet
            {"text": "ขาว แดง เขียว เหลือง ชมพู ม่วง เทา",
             "font": "Niramit",
             "screenshot": "textbox_tyke_2.png"
             }
        ]
        # Test each case and compare against screenshot
        for case in exemplars + tykes:
            textbox.reset()
            textbox.fontMGR.addGoogleFont(case['font'])
            textbox.font = case['font']
            textbox.text = case['text']
            self.win.flip()
            textbox.draw()
            if case['screenshot']:
                # Uncomment to save current configuration as desired
                #self.win.getMovieFrame(buffer='back').save(Path(utils.TESTS_DATA_PATH) / case['screenshot'])
                utils.compareScreenshot(Path(utils.TESTS_DATA_PATH) / case['screenshot'], self.win, crit=20)

    def test_colors(self):
        textbox = TextBox2(self.win, "",
                           "Consolas", pos=(0, 0), size=(1, 1), letterHeight=0.1, units='height', colorSpace="rgb")
        textbox.fontMGR.addGoogleFont("Noto Sans")
        textbox.font = "Noto Sans"
        textbox.text = "A PsychoPy zealot knows a smidge of wx, but JavaScript is the question."
        # Some exemplar text to test basic colors
        exemplars = [
            # White on black in rgb
            {"color": (1, 1, 1), "fillColor": (-1,-1,-1), "borderColor": (-1,-1,-1), "space": "rgb",
             "screenshot": "textbox_colors_WOB.png"},
            # White on black in named
            {"color": "white", "fillColor": "black", "borderColor": "black", "space": "rgb",
             "screenshot": "textbox_colors_WOB.png"},
            # White on black in hex
            {"color": "#ffffff", "fillColor": "#000000", "borderColor": "#000000", "space": "hex",
             "screenshot": "textbox_colors_WOB.png"},
            {"color": "red", "fillColor": "yellow", "borderColor": "blue", "space": "rgb",
             "screenshot": "textbox_colors_exemplar1.png"},
            {"color": "yellow", "fillColor": "blue", "borderColor": "red", "space": "rgb",
             "screenshot": "textbox_colors_exemplar2.png"},
            {"color": "blue", "fillColor": "red", "borderColor": "yellow", "space": "rgb",
             "screenshot": "textbox_colors_exemplar3.png"},
        ]
        # Some colors which are likely to cause problems if something isn't working
        tykes = [
            # Text only
            {"color": "white", "fillColor": None, "borderColor": None, "space": "rgb",
             "screenshot": "textbox_colors_tyke1.png"},
            # Fill only
            {"color": None, "fillColor": "white", "borderColor": None, "space": "rgb",
             "screenshot": "textbox_colors_tyke2.png"},
            # Border only
            {"color": None, "fillColor": None, "borderColor": "white", "space": "rgb",
            "screenshot": "textbox_colors_tyke3.png"},
        ]
        # Test each case and compare against screenshot
        for case in exemplars + tykes:
            # Raise error if case spec does not contain all necessary keys
            if not all(key in case for key in ["color", "fillColor", "borderColor", "space", "screenshot"]):
                raise KeyError(f"Case spec for test_colors in class {self.__class__.__name__} ({__file__}) invalid, test cannot be run.")
            # Apply params from case spec
            textbox.colorSpace = case['space']
            textbox.color = case['color']
            textbox.fillColor = case['fillColor']
            textbox.borderColor = case['borderColor']
            self.win.flip()
            textbox.draw()
            if case['screenshot']:
                # Uncomment to save current configuration as desired
                # self.win.getMovieFrame(buffer='back').save(Path(utils.TESTS_DATA_PATH) / case['screenshot'])
                utils.compareScreenshot(Path(utils.TESTS_DATA_PATH) / case['screenshot'], self.win, crit=20)


    def test_basic(self):
        pass

    def test_something(self):
        # to-do: test visual display, char position, etc
        pass

    def test_font_manager(self):
        # Create a font manager
        mgr = FontManager()
        # Check that it finds fonts which should be pre-packaged with PsychoPy in the resources folder
        assert bool(mgr.getFontNamesSimilar("Open Sans"))
        # Check that it doesn't find fonts which aren't installed as default
        assert not bool(mgr.getFontNamesSimilar("Dancing Script"))
        # Check that it can install fonts from Google
        mgr.addGoogleFont("Hanalei")
        # Check that these fonts are found once installed
        assert bool(mgr.getFontNamesSimilar("Hanalei"))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""
Simple interface to the Cetoni neMESYS syringe pump system, based on the
`pyqmix <https://github.com/psyfood/pyqmix/>`_ library. The syringe pump
system is described in the following publication:

    CA Andersen, L Alfine, K Ohla, & R Höchenberger (2018):
    "A new gustometer: Template for the construction of a portable and
     modular stimulator for taste and lingual touch."
    Behavior Research Methods. doi: 10.3758/s13428-018-1145-1

"""


from psychopy import prefs
from pyqmix import config, QmixBus, QmixPump
import pyqmix.pump

syringeTypes = list(pyqmix.pump.syringes.keys())
volumeUnits = ['mL']
flowRateUnits = ['mL/s', 'mL/min']
configName = prefs.hardware['qmixConfiguration']
bus = None
pumps = []  # To keep track of all instantiated pumps.


def _init_bus():
    global bus

    config.set_qmix_config(configName)
    bus = QmixBus()


def _init_all_pumps():
    # Initialize the very first pump to make its `n_pumps` property visible.
    # After that, instantiate all remaining pumps.
    p = Pump(index=0)
    n_pumps = p._pyqmix_pump.n_pumps
    [Pump(index=index) for index in range(1, n_pumps)]


class Pump(object):
    """
    An interface to Cetoni neMESYS syringe pumps, based on the
    `pyqmix <https://github.com/psyfood/pyqmix/>`_ library.
    """
    def __init__(self, index, volumeUnit='mL',
                 flowRateUnit='mL/s',
                 syringeType='50 mL glass'):
        """

        Parameters
        ----------
        index : int
            The index of the pump. The first pump in the system has `index=0`,
            the second `index=1`, etc.

        volumeUnit: 'mL'
            The unit in which the volumes are provided. Currently, only `'ml'`
            is supported.

        flowRateUnit : 'mL/s' or 'mL/min
            The unit in which flow rates are provided.

        syringeType : '25 mL glass' or '50 mL glass'
            Type of the installed syringe, `as understood by pyqmix
            <https://pyqmix.readthedocs.io/en/latest/interface.html#pyqmix.pump.QmixPump.set_syringe_params_by_type>`_.

        """
        # Only initialize the bus when instantiating the very first pump.
        if bus is None:
            _init_bus()

        self._pyqmix_pump = QmixPump(index=index)
        self.volumeUnit = volumeUnit
        self.flowRateUnit = flowRateUnit
        self.syringeType = syringeType
        self.index = index
        pumps.append(self)

    @property
    def fillLevel(self):
        """Current fill level of the syringe.
        """
        return self._pyqmix_pump.fill_level

    @property
    def maxFlowRate(self):
        """Maximum flow rate the pump can provide with the installed syringe.
        """
        return self._pyqmix_pump.max_flow_rate

    @property
    def isInFaultState(self):
        """Whether the pump is currently in a non-operational "fault state".

        To enable the pump again, call :meth:`~Pump.clearFaultState`.
        """
        return self._pyqmix_pump.is_in_fault_state

    @property
    def syringeType(self):
        """Type of the installed syringe.
        """
        return self._syringeType

    @syringeType.setter
    def syringeType(self, syringeType):
        self._pyqmix_pump.set_syringe_params_by_type(syringeType)
        self._syringeType = syringeType

    @property
    def volumeUnit(self):
        """The unit in which the volumes are provided.
        """
        return self._volumeUnit

    @volumeUnit.setter
    def volumeUnit(self, volumeUnit):
        if volumeUnit not in volumeUnits:
            raise ValueError('Volume unit must be one of %s' % volumeUnits)

        self._pyqmix_pump.set_volume_unit(prefix='milli', unit='litres')
        self._volumeUnit = 'mL'

    @property
    def flowRateUnit(self):
        """The unit in which flow rates are provided.
        """
        return self._flowRateUnit

    @flowRateUnit.setter
    def flowRateUnit(self, flowRateUnit):
        if flowRateUnit == 'mL/s':
            flow_time_unit = 'per_second'
        elif flowRateUnit == 'mL/min':
            flow_time_unit = 'per_minute'
        else:
            msg = 'Invalid flow rate unit: must be one of' %flowRateUnits
            raise ValueError(msg)

        self._pyqmix_pump.set_flow_unit(prefix='milli',
                                        volume_unit='litres',
                                        time_unit=flow_time_unit)
        self._flowRateUnit = flowRateUnit

    def clearFaultState(self):
        """Switch the pump back to an operational state after an error had
        occurred.
        """
        self._pyqmix_pump.clear_fault_state()

    def calibrate(self, waitUntilDone=False):
        """Calibrate the syringe pump.

        You must not use this function if a syringe is installed in the
        pump as the syringe may be damaged!

        Parameters
        ----------
        waitUntilDone : bool
            Whether to block program execution until calibration is completed.

        """
        self._pyqmix_pump.calibrate(wait_until_done=waitUntilDone)

    def dispense(self, volume, flowRate, waitUntilDone=False,
                 switchValveWhenDone=False):
        """Dispense the specified volume.

        Parameters
        ----------
        volume : float
            The volume to dispense.
        flowRate : float
            The desired flow rate.
        waitUntilDone : bool
            Whether to block program execution until calibration is completed.
        switchValveWhenDone : bool
            If `True`, switch the valve to aspiation position after the
            dispense is finished. Implies `wait_until_done=True`.

        """
        if flowRate <= 0 or flowRate > self._pyqmix_pump.max_flow_rate:
            msg = ('Flow rate must be positive and <= .3%f %s'
                   % (self.maxFlowRate,
                      self.flowRateUnit))
            raise ValueError(msg)

        self._pyqmix_pump.dispense(volume=volume,
                                   flow_rate=flowRate,
                                   wait_until_done=waitUntilDone,
                                   switch_valve_when_done=switchValveWhenDone)

    def aspirate(self, volume, flowRate, waitUntilDone=False,
                 switchValveWhenDone=False):
        """Aspirate the specified volume.

        Parameters
        ----------
        volume : float
            The volume to aspirate.
        flowRate : float
            The desired flow rate.
        waitUntilDone : bool
            Whether to block program execution until calibration is completed.
        switchValveWhenDone : bool
            If `True`, switch the valve to dispense position after the
            aspiration is finished. Implies `wait_until_done=True`.

        """
        if flowRate <= 0 or flowRate > self._pyqmix_pump.max_flow_rate:
            msg = ('Flow rate must be positive and <= .3%f %s'
                   % (self.maxFlowRate,
                      self.flowRateUnit))
            raise ValueError(msg)

        self._pyqmix_pump.aspirate(volume=volume,
                                   flow_rate=flowRate,
                                   wait_until_done=waitUntilDone,
                                   switch_valve_when_done=switchValveWhenDone)

    def fill(self, flowRate, waitUntilDone=False,
             switchValveWhenDone=False):
        """Fill the syringe entirely.

        Parameters
        ----------
        flowRate : float
            The desired flow rate.
        waitUntilDone : bool
            Whether to block program execution until calibration is completed.
        switchValveWhenDone : bool
            If `True`, switch the valve to dispense position after the
            aspiration is finished. Implies `wait_until_done=True`.

        """
        if flowRate <= 0 or flowRate > self._pyqmix_pump.max_flow_rate:
            msg = ('Flow rate must be positive and <= .3%f %s'
                   % (self.maxFlowRate,
                      self.flowRateUnit))
            raise ValueError(msg)

        self._pyqmix_pump.fill(flow_rate=flowRate,
                               wait_until_done=waitUntilDone,
                               switch_valve_when_done=switchValveWhenDone)

    def empty(self, flowRate, waitUntilDone=False,
              switchValveWhenDone=False):
        """Empty the syringe entirely.

        Parameters
        ----------
        flowRate : float
            The desired flow rate.
        waitUntilDone : bool
            Whether to block program execution until calibration is completed.
        switchValveWhenDone : bool
            If `True`, switch the valve to aspirate position after the
            dispensing is finished. Implies `wait_until_done=True`.

        """
        if flowRate <= 0 or flowRate > self._pyqmix_pump.max_flow_rate:
            msg = ('Flow rate must be positive and <= .3%f %s'
                   % (self.maxFlowRate,
                      self.flowRateUnit))
            raise ValueError(msg)

        self._pyqmix_pump.empty(flow_rate=flowRate,
                                wait_until_done=waitUntilDone,
                                switch_valve_when_done=switchValveWhenDone)

    def switchValvePosition(self):
        """Switch the valve to the opposite position.
        """
        self._pyqmix_pump.valve.switch_position()

    def stop(self):
        """Stop any pump operation immediately.
        """
        self._pyqmix_pump.stop()


class _PumpWrapperForBuilderComponent(object):
    """
    Merely for use in the corresponding Builder component, to allow
    re-using the same Pump (in different components) within the same
    routine. To make this possible, we expose a `status` attribute,
    allowing e.g. one Pump Component to finish, while another Pump Component
    in the same routine will be left entirely unaffected, even if it is
    referring to the exact same Pump (i.e., same pump index).

    The methods implemented here simply pass on their arguments to the
    respective methods in the `Pump` instance (stored in `.pump`); this just
    happens for convenience: One can now e.g. invoke
    `_PumpWrapperForBuilderComponent.empty()` instead of
    `_PumpWrapperForBuilderComponent.pump.empty()`, which should make the
    code generated by Builder easier to read and work with.

    """
    def __init__(self, pump):
        """
        Parameters
        ----------
        pump : Pump

        """
        self.pump = pump
        self.status = None

    def fill(self, *args, **kwargs):
        self.pump.fill(*args, **kwargs)

    def empty(self, *args, **kwargs):
        self.pump.empty(*args, **kwargs)

    def stop(self):
        self.pump.stop()

    def switchValvePosition(self):
        self.pump.switchValvePosition()

    @property
    def syringeType(self):
        return self.pump.syringeType

    @syringeType.setter
    def syringeType(self, syringeType):
        self.pump.syringeType = syringeType

    @property
    def flowRateUnit(self):
        return self.pump.flowRateUnit

    @flowRateUnit.setter
    def flowRateUnit(self, flowRateUnit):
        self.pump.flowRateUnit = flowRateUnit



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This is an extended version of the base timeByFrames demo, adding
the ability to set the psychopy experiment runtime process priority
and disable python garbage collection to see if either influences
the precision of your frame flips.
"""

from __future__ import division
from __future__ import print_function

from builtins import str
from builtins import range
import gc, numpy
from psychopy import visual, logging, core, event

nIntervals = 500
visual.useFBO = True  # if available (try without for comparison)
disable_gc = False
process_priority = 'normal'  # 'high' or 'realtime'

if process_priority == 'normal':
    pass
elif process_priority == 'high':
    core.rush(True)
elif process_priority == 'realtime':
    # Only makes a diff compared to 'high' on Windows.
    core.rush(True, realtime=True)
else:
    print('Invalid process priority:', process_priority, "Process running at normal.")
    process_priority = 'normal'

if disable_gc:
    gc.disable()

import matplotlib
matplotlib.use('Qt4Agg')  # change this to control the plotting 'back end'
import pylab

win = visual.Window([1280, 1024], fullscr=True, allowGUI=False, waitBlanking=True)
progBar = visual.GratingStim(win, tex=None, mask=None,
    size=[0, 0.05], color='red', pos=[0, -0.9], autoLog=False)
myStim = visual.GratingStim(win, tex='sin', mask='gauss',
    size=300, sf=0.05, units='pix', autoLog=False)

# logging.console.setLevel(logging.INFO)  # uncomment to print log every frame

fliptimes = numpy.zeros(nIntervals + 1)

win.recordFrameIntervals = True
for frameN in range(nIntervals + 1):
    progBar.setSize([2.0 * frameN / nIntervals, 0.05])
    progBar.draw()
    myStim.setPhase(0.1, '+')
    myStim.setOri(1, '+')
    myStim.draw()
    if event.getKeys():
        print('stopped early')
        break
    win.logOnFlip(msg='frame=%i' % frameN, level=logging.EXP)
    fliptimes[frameN] = win.flip()

if disable_gc:
    gc.enable()
core.rush(False)

win.close()

# calculate some values
intervalsMS = pylab.array(win.frameIntervals) * 1000
m = pylab.mean(intervalsMS)
sd = pylab.std(intervalsMS)
# se=sd/pylab.sqrt(len(intervalsMS)) # for CI of the mean

nTotal = len(intervalsMS)
nDropped = sum(intervalsMS > (1.5 * m))
ifis =(fliptimes[1: ]-fliptimes[: -1]) * 1000

# plot the frameintervals
pylab.figure(figsize=[12, 8], )

pylab.subplot2grid((2, 2), (0, 0), colspan=2)
pylab.plot(intervalsMS, '-')
pylab.ylabel('t (ms)')
pylab.xlabel('frame N')
msg = "Dropped/Frames = %i/%i = %.3f%%. Process Priority: %s, GC Disabled: %s"
pylab.title(msg % (nDropped, nTotal, 100 * nDropped/float(nTotal),
                   process_priority, str(disable_gc)), fontsize=12)

pylab.subplot2grid((2, 2), (1, 0))
pylab.hist(intervalsMS, 50, histtype='stepfilled')
pylab.xlabel('t (ms)')
pylab.ylabel('n frames')
msg = "win.frameIntervals\nMean=%.2fms, s.d.=%.2f, 99%%CI(frame)=%.2f-%.2f"
pylab.title(msg % (m, sd, m-2.58 * sd, m + 2.58 * sd), fontsize=12)

pylab.subplot2grid((2, 2), (1, 1))
pylab.hist(ifis, 50, histtype='stepfilled')
pylab.xlabel('t (ms)')
pylab.ylabel('n frames')
msg = "Inter Flip Intervals\nMean=%.2fms, s.d.=%.2f, range=%.2f-%.2f ms"
pylab.title(msg % (ifis.mean(), ifis.std(), ifis.min(), ifis.max()), fontsize=12)

pylab.tight_layout()
pylab.show()

win.close()
core.quit()

# The contents of this file are in the public domain.



"""ioHub Common Eye Tracker Interface"""
 # Part of the psychopy.iohub library.
 # Copyright (C) 2012-2016 iSolver Software Solutions
 # Distributed under the terms of the GNU General Public License (GPL).

from .. import DeviceEvent
from ...constants import EventConstants
from . import EyeTrackerDevice
import numpy as np

##################### Eye Tracker Sample Stream Types ####################
#


class EyeTrackerEvent(DeviceEvent):
    PARENT_DEVICE = EyeTrackerDevice


class MonocularEyeSampleEvent(EyeTrackerEvent):
    """A MonocularEyeSampleEvent represents the eye position and eye attribute
    data collected from one frame or reading of an eye tracker device that is
    recoding from only one eye, or is recording from both eyes and averaging
    the binocular data.
    
    Event Type ID: EventConstants.MONOCULAR_EYE_SAMPLE

    Event Type String: 'MONOCULAR_EYE_SAMPLE'
    """
    _newDataTypes = [
        # The eye type that the sample is from. Valid values are:
        ('eye', 'u1'),
        #   EyeTrackerConstants.LEFT_EYE
        #   EyeTrackerConstants.RIGHT_EYE
        #   EyeTrackerConstants.BINOCULAR
        #   EyeTrackerConstants.BINOCULAR_AVERAGED
        #   EyeTrackerConstants.SIMULATED_MONOCULAR
        #   EyeTrackerConstants.SIMULATED_BINOCULAR

        # The calibrated horizontal eye position on the calibration plane.
        ('gaze_x', 'f4'),
        # This value is specified in Display Coordinate Type Units.

        # The calibrated vertical eye position on the calibration plane.
        ('gaze_y', 'f4'),
        # This value is specified in Display Coordinate Type Units.

        # The calculated point of gaze in depth. Generally  This can only be
        ('gaze_z', 'f4'),
        # provided if binocular reporting is being performed.

        # The x eye position in an eye trackers 3D coordinate space.
        ('eye_cam_x', 'f4'),
        # Generally this field is only available by systems that are also
        # calculating eye data using a 3D model of eye position relative to
        # the eye camera(s) for example.

        # The y eye position in an eye trackers 3D coordinate space.
        ('eye_cam_y', 'f4'),
        # Generally this field is only available by systems that are also
        # calculating eye data using a 3D model of eye position relative to
        # the eye camera(s) for example.

        # The z eye position in an eye trackers 3D coordinate space.
        ('eye_cam_z', 'f4'),
        # Generally this field is only available by systems that are also
        # calculating eye data using a 3D model of eye position relative to
        # the eye camera(s) for example.

        # The horizontal angle of eye the relative to the head.
        ('angle_x', 'f4'),

        # The vertical angle of eye the relative to the head.
        ('angle_y', 'f4'),

        # The non-calibrated x position of the calculated eye 'center'
        ('raw_x', 'f4'),
        # on the camera sensor image,
        # factoring in any corneal reflection adjustments.
        # This is typically reported in some arbitrary unit space that
        # often has sub-pixel resolution due to image processing techniques
        # being applied.

        # The non-calibrated y position of the calculated eye 'center'
        ('raw_y', 'f4'),
        # on the camera sensor image,
        # factoring in any corneal reflection adjustments.
        # This is typically reported in some arbitrary unit space that
        # often has sub-pixel resolution due to image processing techniques
        # being applied.

        # A measure related to pupil size or diameter. Attribute
        ('pupil_measure1', 'f4'),
        # pupil_measure1_type defines what type the measure represents.

        # Several possible pupil_measure types available:
        ('pupil_measure1_type', 'u1'),
        #       EyeTrackerConstants.PUPIL_AREA
        #       EyeTrackerConstants.PUPIL_DIAMETER
        #       EyeTrackerConstants.PUPIL_WIDTH
        #       EyeTrackerConstants.PUPIL_HEIGHT
        #       EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #       EyeTrackerConstants.PUPIL_MINOR_AXIS

        # A measure related to pupil size or diameter. Attribute
        ('pupil_measure2', 'f4'),
        # pupil_measure2_type defines what type the measure represents.

        # Several possible pupil_measure types are available:
        ('pupil_measure2_type', 'u1'),
        #       EyeTrackerConstants.PUPIL_AREA
        #       EyeTrackerConstants.PUPIL_DIAMETER
        #       EyeTrackerConstants.PUPIL_WIDTH
        #       EyeTrackerConstants.PUPIL_HEIGHT
        #       EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #       EyeTrackerConstants.PUPIL_MINOR_AXIS

        ('ppd_x', 'f4'),     # Horizontal pixels per visual degree for this eye position
        # as reported by the eye tracker.

        ('ppd_y', 'f4'),     # Vertical pixels per visual degree for this eye position
        # as reported by the eye tracker.

        # Horizontal velocity of the eye at the time of the sample;
        ('velocity_x', 'f4'),
        # as reported by the eye tracker.

        # Vertical velocity of the eye at the time of the sample;
        ('velocity_y', 'f4'),
        # as reported by the eye tracker.

        ('velocity_xy', 'f4'),  # 2D Velocity of the eye at the time of the sample;
        # as reported by the eye tracker.

        # An available status word for the eye tracker sample.
        ('status', 'u1')
        # Meaning is completely tracker dependent.
    ]
    EVENT_TYPE_ID = EventConstants.MONOCULAR_EYE_SAMPLE
    EVENT_TYPE_STRING = 'MONOCULAR_EYE_SAMPLE'
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        #: The eye type that the sample is from. Valid values are:
        #:
        #: EventConstants.LEFT_EYE
        #: EventConstants.RIGHT_EYE
        #: EventConstants.SIMULATED_MONOCULAR
        #: EventConstants.MONOCULAR
        self.eye = None

        #: The calibrated horizontal eye position on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.gaze_x = None

        #: The calibrated vertical eye position on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.gaze_y = None

        #: The calculated point of gaze in depth. Generally this can only be
        #: provided if binocular reporting is being performed.
        self.gaze_z = None

        #: The x eye position in an eye trackers 3D coordinate space.
        #: Generally this field is only available by systems that are also
        #: calculating eye data using a 3D model of eye position relative to
        #: the eye camera(s) for example.
        self.eye_cam_x = None

        #: The y eye position in an eye trackers 3D coordinate space.
        #: Generally this field is only available by systems that are also
        #: calculating eye data using a 3D model of eye position relative to
        #: the eye camera(s) for example.
        self.eye_cam_y = None

        #: The z eye position in an eye trackers 3D coordinate space.
        #: Generally this field is only available by systems that are also
        #: calculating eye data using a 3D model of eye position relative to
        #: the eye camera(s) for example.
        self.eye_cam_z = None

        #: The horizontal angle of eye the relative to the head.
        self.angle_x = None

        #: The vertical angle of eye the relative to the head.
        self.angle_y = None

        #: The non-calibrated x position of the calculated eye 'center'
        #: on the camera sensor image, factoring in any corneal reflection
        #: or other low level adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.raw_x = None

        #: The non-calibrated y position of the calculated eye 'center'
        #: on the camera sensor image, factoring in any corneal reflection
        #: or other low level adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.raw_y = None

        #: A measure related to pupil size or diameter. The attribute
        #: pupil_measure1_type defines what type the measure represents.
        self.pupil_measure1 = None

        #: The type of pupil size or shape information provided in the pupil_measure1
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.pupil_measure1_type = None

        #: A second measure related to pupil size or diameter. The attribute
        #: pupil_measure2_type defines what type the measure represents.
        self.pupil_measure2 = None

        #: The type of pupil size or shape information provided in the pupil_measure2
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.pupil_measure2_type = None

        #: Horizontal pixels per visual degree for this eye position
        #: as reported by the eye tracker.
        self.ppd_x = None

        #: Vertical pixels per visual degree for this eye position
        #: as reported by the eye tracker.
        self.ppd_y = None

        #: Horizontal velocity of the eye at the time of the sample;
        #: as reported by the eye tracker.
        self.velocity_x = None

        #: Vertical velocity of the eye at the time of the sample;
        #: as reported by the eye tracker.
        self.velocity_y = None

        #: 2D Velocity of the eye at the time of the sample;
        #: as reported by the eye tracker.
        self.velocity_xy = None

        #: An available status byte for the eye tracker sample.
        #: Meaning is completely tracker dependent.
        self.status = None

        DeviceEvent.__init__(self, *args, **kwargs)


class EyeSampleEvent(EyeTrackerEvent):
    """A EyeSampleEvent reports minimal data regarding an eye sample,
    containing a subset of the full MonocularEyeSampleEvent fields. Support for
    this event type is optional. If requested but not supported, no events of
    this type will be returned.

    Both EyeSampleEvents and MonocularEyeSampleEvent /
    BinocularEyeSampleEvents can be requested from the eye tracker,
    although doing so is redundant and therefore not suggested. ;)

    If binocular eye data is available from the device sample, left and right
    eye data will be combined to provide a single x,y position and pupil size.
    How the binocular eye data is combined into a EyeSampleEvent is
    determined by each eye tracker interface.

    Please refer to the implementation specific documentation for the
    eye tracker of interest for more details.

    Event Type ID: EventConstants.EYE_SAMPLE

    Event Type String: 'EYE_SAMPLE'

    """
    _newDataTypes = [
        ('x', np.float32),  # The horizontal eye position. Unit type used is
        # implementation specific.

        ('y', np.float32),  # The vertical eye position. Unit type used is
        # implementation specific.

        ('pupil', np.float32),  # Pupil size or diameter.
        # Unit type used is implementation specific.

        ('frame', np.uint64),  # Device frame number for the sample.

        ('status', np.uint32)  # Status of the eye tracker sample.
    ]

    EVENT_TYPE_ID = EventConstants.EYE_SAMPLE
    EVENT_TYPE_STRING = 'EYE_SAMPLE'
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        #: The horizontal eye position.
        self.x = None

        #: The vertical eye position.
        self.y = None

        #: Pupil size or diameter.
        self.pupil = None

        #: Device frame number for the sample.
        self.frame = None

        #: An available status byte for the eye tracker sample.
        #: Meaning is completely tracker dependent.
        self.status = None

        DeviceEvent.__init__(self, *args, **kwargs)


class BinocularEyeSampleEvent(EyeTrackerEvent):
    """The BinocularEyeSampleEvent event represents the eye position and eye
    attribute data collected from one frame or reading of an eye tracker device
    that is recording both eyes of a participant.

    Event Type ID: EventConstants.BINOCULAR_EYE_SAMPLE

    Event Type String: 'BINOCULAR_EYE_SAMPLE'
    """
    _newDataTypes = [
        ('left_gaze_x', 'f4'),
        ('left_gaze_y', 'f4'),
        ('left_gaze_z', 'f4'),
        ('left_eye_cam_x', 'f4'),
        ('left_eye_cam_y', 'f4'),
        ('left_eye_cam_z', 'f4'),
        ('left_angle_x', 'f4'),
        ('left_angle_y', 'f4'),
        ('left_raw_x', 'f4'),
        ('left_raw_y', 'f4'),
        ('left_pupil_measure1', 'f4'),
        ('left_pupil_measure1_type', 'u1'),
        ('left_pupil_measure2', 'f4'),
        ('left_pupil_measure2_type', 'u1'),
        ('left_ppd_x', 'f4'),
        ('left_ppd_y', 'f4'),
        ('left_velocity_x', 'f4'),
        ('left_velocity_y', 'f4'),
        ('left_velocity_xy', 'f4'),
        ('right_gaze_x', 'f4'),
        ('right_gaze_y', 'f4'),
        ('right_gaze_z', 'f4'),
        ('right_eye_cam_x', 'f4'),
        ('right_eye_cam_y', 'f4'),
        ('right_eye_cam_z', 'f4'),
        ('right_angle_x', 'f4'),
        ('right_angle_y', 'f4'),
        ('right_raw_x', 'f4'),
        ('right_raw_y', 'f4'),
        ('right_pupil_measure1', 'f4'),
        ('right_pupil_measure1_type', 'u1'),
        ('right_pupil_measure2', 'f4'),
        ('right_pupil_measure2_type', 'u1'),
        ('right_ppd_x', 'f4'),
        ('right_ppd_y', 'f4'),
        ('right_velocity_x', 'f4'),
        ('right_velocity_y', 'f4'),
        ('right_velocity_xy', 'f4'),
        ('status', 'u1')
    ]

    EVENT_TYPE_ID = EventConstants.BINOCULAR_EYE_SAMPLE
    EVENT_TYPE_STRING = 'BINOCULAR_EYE_SAMPLE'
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):

        #: The calibrated horizontal left eye position on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.left_gaze_x = None

        #: The calibrated vertical left eye position on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.left_gaze_y = None

        #: The calculated point of gaze in depth. Generally this can only be
        #: provided if binocular reporting is being performed.
        self.left_gaze_z = None

        #: The x left eye position in an eye trackers 3D coordinate space.
        #: Generally this field is only available by systems that are also
        #: calculating eye data using a 3D model of eye position relative to
        #: the eye camera(s) for example.
        self.left_eye_cam_x = None

        #: The y left eye position in an eye trackers 3D coordinate space.
        #: Generally this field is only available by systems that are also
        #: calculating eye data using a 3D model of eye position relative to
        #: the eye camera(s) for example.
        self.left_eye_cam_y = None

        #: The z left eye position in an eye trackers 3D coordinate space.
        #: Generally this field is only available by systems that are also
        #: calculating eye data using a 3D model of eye position relative to
        #: the eye camera(s) for example.
        self.left_eye_cam_z = None

        #: The horizontal angle of left eye the relative to the head.
        self.left_angle_x = None

        #: The vertical angle of left eye the relative to the head.
        self.left_angle_y = None

        #: The non-calibrated x position of the calculated left eye 'center'
        #: on the camera sensor image,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.left_raw_x = None

        #: The non-calibrated y position of the calculated left eye 'center'
        #: on the camera sensor image,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.left_raw_y = None

        #: A measure related to left pupil size or diameter. The attribute
        #: pupil_measure1_type defines what type the measure represents.
        self.left_pupil_measure1 = None

        #: The type of left pupil size or shape information provided in the pupil_measure1
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.left_pupil_measure1_type = None

        #: A second measure related to left pupil size or diameter. The attribute
        #: pupil_measure2_type defines what type the measure represents.
        self.left_pupil_measure2 = None

        #: The type of left pupil size or shape information provided in the pupil_measure2
        #: attribute. Several possible pupil_measure types available:
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.left_pupil_measure2_type = None

        #: Horizontal pixels per visual degree for this left eye position
        #: as reported by the eye tracker.
        self.left_ppd_x = None

        #: Vertical pixels per visual degree for this left eye position
        #: as reported by the eye tracker.
        self.left_ppd_y = None

        #: Horizontal velocity of the left eye at the time of the sample;
        #: as reported by the eye tracker.
        self.left_velocity_x = None

        #: Vertical velocity of the left eye at the time of the sample;
        #: as reported by the eye tracker.
        self.left_velocity_y = None

        #: 2D Velocity of the left eye at the time of the sample;
        #: as reported by the eye tracker.
        self.left_velocity_xy = None

        #: The calibrated horizontal right eye position on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.right_gaze_x = None

        #: The calibrated vertical right eye position on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.right_gaze_y = None

        #: The calculated point of gaze in depth. Generally this can only be
        #: provided if binocular reporting is being performed.
        self.right_gaze_z = None

        #: The x right eye position in an eye trackers 3D coordinate space.
        #: Generally this field is only available by systems that are also
        #: calculating eye data using a 3D model of eye position relative to
        #: the eye camera(s) for example.
        self.right_eye_cam_x = None

        #: The y right eye position in an eye trackers 3D coordinate space.
        #: Generally this field is only available by systems that are also
        #: calculating eye data using a 3D model of eye position relative to
        #: the eye camera(s) for example.
        self.right_eye_cam_y = None

        #: The z right eye position in an eye trackers 3D coordinate space.
        #: Generally this field is only available by systems that are also
        #: calculating eye data using a 3D model of eye position relative to
        #: the eye camera(s) for example.
        self.right_eye_cam_z = None

        #: The horizontal angle of right eye the relative to the head.
        self.right_angle_x = None

        #: The vertical angle of right eye the relative to the head.
        self.right_angle_y = None

        #: The non-calibrated x position of the calculated right eye 'center'
        #: on the camera sensor image,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.right_raw_x = None

        #: The non-calibrated y position of the calculated right eye 'center'
        #: on the camera sensor image,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.right_raw_y = None

        #: A measure related to right pupil size or diameter. The attribute
        #: pupil_measure1_type defines what type the measure represents.
        self.right_pupil_measure1 = None

        #: The type of right pupil size or shape information provided in the pupil_measure1
        #: attribute. Several possible pupil_measure types available:
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.right_pupil_measure1_type = None

        #: A second measure related to right pupil size or diameter. The attribute
        #: pupil_measure2_type defines what type the measure represents.
        self.right_pupil_measure2 = None

        #: The type of right pupil size or shape information provided in the pupil_measure2
        #: attribute. Several possible pupil_measure types available:
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.right_pupil_measure2_type = None

        #: Horizontal pixels per visual degree for this right eye position
        #: as reported by the eye tracker.
        self.right_ppd_x = None

        #: Vertical pixels per visual degree for this right eye position
        #: as reported by the eye tracker.
        self.right_ppd_y = None

        #: Horizontal velocity of the right eye at the time of the sample;
        #: as reported by the eye tracker.
        self.right_velocity_x = None

        #: Vertical velocity of the right eye at the time of the sample;
        #: as reported by the eye tracker.
        self.right_velocity_y = None

        #: 2D Velocity of the right eye at the time of the sample;
        #: as reported by the eye tracker.
        self.right_velocity_xy = None

        #: An available status byte for the eye tracker sample.
        #: Meaning is completely tracker dependent.
        self.status = None

        DeviceEvent.__init__(self, *args, **kwargs)

#
################### Fixation Event Types ##########################
#


class FixationStartEvent(EyeTrackerEvent):
    """A FixationStartEvent is generated when the beginning of an eye fixation
    ( in very general terms, a period of relatively stable eye position ) is
    detected by the eye trackers sample parsing algorithms.

    Event Type ID: EventConstants.FIXATION_START

    Event Type String: 'FIXATION_START'
    
    """
    _newDataTypes = [
        # The eye type that the fixation is from. Valid values are:
        ('eye', 'u1'),
        #   EyeTrackerConstants.LEFT_EYE
        #   EyeTrackerConstants.RIGHT_EYE
        #   EyeTrackerConstants.BINOCULAR_AVERAGED
        #   EyeTrackerConstants.SIMULATED_MONOCULAR

        # The calibrated horizontal eye position on the calibration plane.
        ('gaze_x', 'f4'),
        # This value is specified in Display Coordinate Type Units.

        # The calibrated vertical eye position on the calibration plane.
        ('gaze_y', 'f4'),
        # This value is specified in Display Coordinate Type Units.

        # The calculated point of gaze in depth. Generally  This can only be
        ('gaze_z', 'f4'),
        # provided if binocular reporting is being performed.

        # The horizontal angle of eye the relative to the head.
        ('angle_x', 'f4'),

        # The vertical angle of eye the relative to the head.
        ('angle_y', 'f4'),

        # The non-calibrated x position of the calculated eye 'center'
        ('raw_x', 'f4'),
        # on the camera sensor image,
        # factoring in any corneal reflection adjustments.
        # This is typically reported in some arbitrary unit space that
        # often has sub-pixel resolution due to image processing techniques
        # being applied.

        # The non-calibrated y position of the calculated eye 'center'
        ('raw_y', 'f4'),
        # on the camera sensor image,
        # factoring in any corneal reflection adjustments.
        # This is typically reported in some arbitrary unit space that
        # often has sub-pixel resolution due to image processing techniques
        # being applied.

        # A measure related to pupil size or diameter. Attribute
        ('pupil_measure1', 'f4'),
        # pupil_measure1_type defines what type the measure represents.

        # Several possible pupil_measure types available:
        ('pupil_measure1_type', 'u1'),
        #       EyeTrackerConstants.PUPIL_AREA
        #       EyeTrackerConstants.PUPIL_DIAMETER
        #       EyeTrackerConstants.PUPIL_WIDTH
        #       EyeTrackerConstants.PUPIL_HEIGHT
        #       EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #       EyeTrackerConstants.PUPIL_MINOR_AXIS

        # A measure related to pupil size or diameter. Attribute
        ('pupil_measure2', 'f4'),
        # pupil_measure2_type defines what type the measure represents.

        # Several possible pupil_measure types are available:
        ('pupil_measure2_type', 'u1'),
        #       EyeTrackerConstants.PUPIL_AREA
        #       EyeTrackerConstants.PUPIL_DIAMETER
        #       EyeTrackerConstants.PUPIL_WIDTH
        #       EyeTrackerConstants.PUPIL_HEIGHT
        #       EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #       EyeTrackerConstants.PUPIL_MINOR_AXIS

        ('ppd_x', 'f4'),     # Horizontal pixels per visual degree for this eye position
        # as reported by the eye tracker.

        ('ppd_y', 'f4'),     # Vertical pixels per visual degree for this eye position
        # as reported by the eye tracker.

        # Horizontal velocity of the eye at the time of the fixation start
        # sample;
        ('velocity_x', 'f4'),
        # as reported by the eye tracker.

        # Vertical velocity of the eye at the time of the fixation start
        # sample;
        ('velocity_y', 'f4'),
        # as reported by the eye tracker.

        # 2D Velocity of the eye at the time of the fixation start sample;
        ('velocity_xy', 'f4'),
        # as reported by the eye tracker.

        # An available status word for the eye tracker fixation start event.
        ('status', 'u1')
        # Meaning is completely tracker dependent.
    ]

    EVENT_TYPE_ID = EventConstants.FIXATION_START
    EVENT_TYPE_STRING = EventConstants.getName(EVENT_TYPE_ID)
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):

        #: The eye type that the event is from. Valid values are:
        #: EyeTrackerConstants.LEFT_EYE
        #: EyeTrackerConstants.RIGHT_EYE
        #: EyeTrackerConstants.MONOCULAR
        #: EyeTrackerConstants.SIMULATED_MONOCULAR
        self.eye = None

        #: The calibrated horizontal eye position at the start of the eye event
        #: on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.gaze_x = None

        #: The calibrated vertical eye position at the start of the eye event on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.gaze_y = None

        #: The calculated point of gaze in depth at the start of the eye event. Generally this can only be
        #: provided if binocular reporting is being performed.
        self.gaze_z = None

        #: The horizontal angle of eye the relative to the head at the start of the eye event.
        self.angle_x = None

        #: The vertical angle of eye the relative to the head at the start of the eye event.
        self.angle_y = None

        #: The non-calibrated x position of the calculated eye 'center'
        #: on the camera sensor image at the start of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.raw_x = None

        #: The non-calibrated y position of the calculated eye 'center'
        #: on the camera sensor image at the start of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.raw_y = None

        #: A measure related to pupil size or diameter at the start of the eye event.
        #: The attribute pupil_measure1_type defines what type the measure represents.
        self.pupil_measure1 = None

        #: The type of pupil size or shape information provided in the pupil_measure1
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.pupil_measure1_type = None

        #: A second measure related to pupil size or diameter at the start of the eye event.
        #: The attribute pupil_measure2_type defines what type the measure represents.
        self.pupil_measure2 = None

        #: The type of pupil size or shape information provided in the pupil_measure2
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.pupil_measure2_type = None

        #: Horizontal pixels per visual degree for this eye position at the start of the eye event
        #: as reported by the eye tracker.
        self.ppd_x = None

        #: Vertical pixels per visual degree for this eye position at the start of the eye event
        #: as reported by the eye tracker.
        self.ppd_y = None

        #: Horizontal velocity of the eye at the start of the eye event;
        #: as reported by the eye tracker.
        self.velocity_x = None

        #: Vertical velocity of the eye at the start of the eye event;
        #: as reported by the eye tracker.
        self.velocity_y = None

        #: 2D Velocity of the eye at the start of the eye event;
        #: as reported by the eye tracker.
        self.velocity_xy = None

        #: An available status byte for the eye tracker event.
        #: Meaning or use is completely tracker dependent.
        self.status = None

        DeviceEvent.__init__(self, *args, **kwargs)


class FixationEndEvent(EyeTrackerEvent):
    """A FixationEndEvent is generated when the end of an eye fixation
    ( in very general terms, a period of relatively stable eye position ) is
    detected by the eye trackers sample parsing algorithms.

    Event Type ID: EventConstants.FIXATION_END

    Event Type String: 'FIXATION_END'

    """
    _newDataTypes = [
        ('eye', 'u1'),
        ('duration', 'f4'),
        ('start_gaze_x', 'f4'),
        ('start_gaze_y', 'f4'),
        ('start_gaze_z', 'f4'),
        ('start_angle_x', 'f4'),
        ('start_angle_y', 'f4'),
        ('start_raw_x', 'f4'),
        ('start_raw_y', 'f4'),
        ('start_pupil_measure1', 'f4'),
        ('start_pupil_measure1_type', 'u1'),
        ('start_pupil_measure2', 'f4'),
        ('start_pupil_measure2_type', 'u1'),
        ('start_ppd_x', 'f4'),
        ('start_ppd_y', 'f4'),
        ('start_velocity_x', 'f4'),
        ('start_velocity_y', 'f4'),
        ('start_velocity_xy', 'f4'),
        ('end_gaze_x', 'f4'),
        ('end_gaze_y', 'f4'),
        ('end_gaze_z', 'f4'),
        ('end_angle_x', 'f4'),
        ('end_angle_y', 'f4'),
        ('end_raw_x', 'f4'),
        ('end_raw_y', 'f4'),
        ('end_pupil_measure1', 'f4'),
        ('end_pupil_measure1_type', 'u1'),
        ('end_pupil_measure2', 'f4'),
        ('end_pupil_measure2_type', 'u1'),
        ('end_ppd_x', 'f4'),
        ('end_ppd_y', 'f4'),
        ('end_velocity_x', 'f4'),
        ('end_velocity_y', 'f4'),
        ('end_velocity_xy', 'f4'),
        ('average_gaze_x', 'f4'),
        ('average_gaze_y', 'f4'),
        ('average_gaze_z', 'f4'),
        ('average_angle_x', 'f4'),
        ('average_angle_y', 'f4'),
        ('average_raw_x', 'f4'),
        ('average_raw_y', 'f4'),
        ('average_pupil_measure1', 'f4'),
        ('average_pupil_measure1_type', 'u1'),
        ('average_pupil_measure2', 'f4'),
        ('average_pupil_measure2_type', 'u1'),
        ('average_ppd_x', 'f4'),
        ('average_ppd_y', 'f4'),
        ('average_velocity_x', 'f4'),
        ('average_velocity_y', 'f4'),
        ('average_velocity_xy', 'f4'),
        ('peak_velocity_x', 'f4'),
        ('peak_velocity_y', 'f4'),
        ('peak_velocity_xy', 'f4'),
        ('status', 'u1')
    ]

    EVENT_TYPE_ID = EventConstants.FIXATION_END
    EVENT_TYPE_STRING = EventConstants.getName(EVENT_TYPE_ID)
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):

        #: The eye type that the event is from. Valid values are:
        #: The eye type that the sample is from. Valid values are:
        #:
        #: EventConstants.LEFT_EYE
        #: EventConstants.RIGHT_EYE
        #: EventConstants.SIMULATED_MONOCULAR
        #: EventConstants.MONOCULAR
        self.eye = None

        #: The calculated duration of the Eye event in sec.msec-usec
        #: format.
        self.duration = None

        #: The calibrated horizontal eye position at the start of the eye event
        #: on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.start_gaze_x = None

        #: The calibrated vertical eye position at the start of the eye event on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.start_gaze_y = None

        #: The calculated point of gaze in depth at the start of the eye event. Generally this can only be
        #: provided if binocular reporting is being performed.
        self.start_gaze_z = None

        #: The horizontal angle of eye the relative to the head at the start of the eye event.
        self.start_angle_x = None

        #: The vertical angle of eye the relative to the head at the start of the eye event.
        self.start_angle_y = None

        #: The non-calibrated x position of the calculated eye 'center'
        #: on the camera sensor image at the start of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.start_raw_x = None

        #: The non-calibrated y position of the calculated eye 'center'
        #: on the camera sensor image at the start of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.start_raw_y = None

        #: A measure related to pupil size or diameter at the start of the eye event.
        #: The attribute pupil_measure1_type defines what type the measure represents.
        self.start_pupil_measure1 = None

        #: The type of pupil size or shape information provided in the pupil_measure1
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.start_pupil_measure1_type = None

        #: A second measure related to pupil size or diameter at the start of the eye event.
        #: The attribute pupil_measure2_type defines what type the measure represents.
        self.start_pupil_measure2 = None

        #: The type of pupil size or shape information provided in the pupil_measure2
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.start_pupil_measure2_type = None

        #: Horizontal pixels per visual degree for this eye position at the start of the eye event
        #: as reported by the eye tracker.
        self.start_ppd_x = None

        #: Vertical pixels per visual degree for this eye position at the start of the eye event
        #: as reported by the eye tracker.
        self.start_ppd_y = None

        #: Horizontal velocity of the eye at the start of the eye event;
        #: as reported by the eye tracker.
        self.start_velocity_x = None

        #: Vertical velocity of the eye at the start of the eye event;
        #: as reported by the eye tracker.
        self.start_velocity_y = None

        #: 2D Velocity of the eye at the start of the eye event;
        #: as reported by the eye tracker.
        self.start_velocity_xy = None

        #: The calibrated horizontal eye position at the end of the eye event
        #: on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.end_gaze_x = None

        #: The calibrated vertical eye position at the end of the eye event on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.end_gaze_y = None

        #: The calculated point of gaze in depth at the end of the eye event. Generally this can only be
        #: provided if binocular reporting is being performed.
        self.end_gaze_z = None

        #: The horizontal angle of eye the relative to the head at the end of the eye event.
        self.end_angle_x = None

        #: The vertical angle of eye the relative to the head at the end of the eye event.
        self.end_angle_y = None

        #: The non-calibrated x position of the calculated eye 'center'
        #: on the camera sensor image at the end of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.end_raw_x = None

        #: The non-calibrated y position of the calculated eye 'center'
        #: on the camera sensor image at the end of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.end_raw_y = None

        #: A measure related to pupil size or diameter at the end of the eye event.
        #: The attribute pupil_measure1_type defines what type the measure represents.
        self.end_pupil_measure1 = None

        #: The type of pupil size or shape information provided in the pupil_measure1
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.end_pupil_measure1_type = None

        #: A second measure related to pupil size or diameter at the end of the eye event.
        #: The attribute pupil_measure2_type defines what type the measure represents.
        self.end_pupil_measure2 = None

        #: The type of pupil size or shape information provided in the pupil_measure2
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.end_pupil_measure2_type = None

        #: Horizontal pixels per visual degree for this eye position at the end of the eye event
        #: as reported by the eye tracker.
        self.end_ppd_x = None

        #: Vertical pixels per visual degree for this eye position at the end of the eye event
        #: as reported by the eye tracker.
        self.end_ppd_y = None

        #: Horizontal velocity of the eye at the end of the eye event;
        #: as reported by the eye tracker.
        self.end_velocity_x = None

        #: Vertical velocity of the eye at the end of the eye event;
        #: as reported by the eye tracker.
        self.end_velocity_y = None

        #: 2D Velocity of the eye at the end of the eye event;
        #: as reported by the eye tracker.
        self.end_velocity_xy = None

        #: Average calibrated horizontal eye position during the eye event
        #: on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.average_gaze_x = None

        #: Average calibrated vertical eye position during the eye event on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.average_gaze_y = None

        #: Average calculated point of gaze in depth during the eye event. Generally this can only be
        #: provided if binocular reporting is being performed.
        self.average_gaze_z = None

        #: Average horizontal angle of eye the relative to the head during the eye event.
        self.average_angle_x = None

        #: Average vertical angle of eye the relative to the head during the eye event.
        self.average_angle_y = None

        #: Average non-calibrated x position of the calculated eye 'center'
        #: on the camera sensor image during the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.average_raw_x = None

        #: The average non-calibrated y position of the calculated eye 'center'
        #: on the camera sensor image during the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.average_raw_y = None

        #: A measure related to average pupil size or diameter during the eye event.
        #: The attribute pupil_measure1_type defines what type the measure represents.
        self.average_pupil_measure1 = None

        #: The type of pupil size or shape information provided in the pupil_measure1
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.average_pupil_measure1_type = None

        #: A second measure related to average pupil size or diameter during the eye event.
        #: The attribute pupil_measure2_type defines what type the measure represents.
        self.average_pupil_measure2 = None

        #: The type of pupil size or shape information provided in the pupil_measure2
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.average_pupil_measure2_type = None

        #: Average Horizontal pixels per visual degree for this eye position during the eye event
        #: as reported by the eye tracker.
        self.average_ppd_x = None

        #: Average Vertical pixels per visual degree for this eye position during the eye event
        #: as reported by the eye tracker.
        self.average_ppd_y = None

        #: Average Horizontal velocity of the eye during the eye event;
        #: as reported by the eye tracker.
        self.average_velocity_x = None

        #: Average Vertical velocity of the eye during the eye event;
        #: as reported by the eye tracker.
        self.average_velocity_y = None

        #: Average 2D Velocity of the eye at the during the eye event;
        #: as reported by the eye tracker.
        self.average_velocity_xy = None

        #: Peak Horizontal velocity of the eye during the eye event;
        #: as reported by the eye tracker.
        self.peak_velocity_x = None

        #: Peak Vertical velocity of the eye during the eye event;
        #: as reported by the eye tracker.
        self.peak_velocity_y = None

        #: Peak 2D Velocity of the eye at the during the eye event;
        #: as reported by the eye tracker.
        self.peak_velocity_xy = None
        
        #: An available status byte for the eye tracker event.
        #: Meaning or use is completely tracker dependent.
        self.status = None

        DeviceEvent.__init__(self, *args, **kwargs)


################### Saccade Event Types ##########################
#
class SaccadeStartEvent(FixationStartEvent):

    EVENT_TYPE_ID = EventConstants.SACCADE_START
    EVENT_TYPE_STRING = EventConstants.getName(EVENT_TYPE_ID)
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING
    __slots__ = []

    def __init__(self, *args, **kwargs):
        FixationStartEvent.__init__(self, *args, **kwargs)


class SaccadeEndEvent(EyeTrackerEvent):
    _newDataTypes = [
        ('eye', 'u1'),
        ('duration', 'f4'),
        ('amplitude_x', 'f4'),
        ('amplitude_y', 'f4'),
        ('angle', 'f4'),
        ('start_gaze_x', 'f4'),
        ('start_gaze_y', 'f4'),
        ('start_gaze_z', 'f4'),
        ('start_angle_x', 'f4'),
        ('start_angle_y', 'f4'),
        ('start_raw_x', 'f4'),
        ('start_raw_y', 'f4'),
        ('start_pupil_measure1', 'f4'),
        ('start_pupil_measure1_type', 'u1'),
        ('start_pupil_measure2', 'f4'),
        ('start_pupil_measure2_type', 'f4'),
        ('start_ppd_x', 'f4'),
        ('start_ppd_y', 'f4'),
        ('start_velocity_x', 'f4'),
        ('start_velocity_y', 'f4'),
        ('start_velocity_xy', 'f4'),
        ('end_gaze_x', 'f4'),
        ('end_gaze_y', 'f4'),
        ('end_gaze_z', 'f4'),
        ('end_angle_x', 'f4'),
        ('end_angle_y', 'f4'),
        ('end_raw_x', 'f4'),
        ('end_raw_y', 'f4'),
        ('end_pupil_measure1', 'f4'),
        ('end_pupil_measure1_type', 'u1'),
        ('end_pupil_measure2', 'f4'),
        ('end_pupil_measure2_type', 'u1'),
        ('end_ppd_x', 'f4'),
        ('end_ppd_y', 'f4'),
        ('end_velocity_x', 'f4'),
        ('end_velocity_y', 'f4'),
        ('end_velocity_xy', 'f4'),
        ('average_velocity_x', 'f4'),
        ('average_velocity_y', 'f4'),
        ('average_velocity_xy', 'f4'),
        ('peak_velocity_x', 'f4'),
        ('peak_velocity_y', 'f4'),
        ('peak_velocity_xy', 'f4'),
        ('status', 'u1')
    ]

    EVENT_TYPE_ID = EventConstants.SACCADE_END
    EVENT_TYPE_STRING = EventConstants.getName(EVENT_TYPE_ID)
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):

        #: The eye type that the event is from. Valid values are:
        #: EyeTrackerConstants.LEFT_EYE
        #: EyeTrackerConstants.RIGHT_EYE
        #: EyeTrackerConstants.MONOCULAR
        #: EyeTrackerConstants.SIMULATED_MONOCULAR
        self.eye = None

        #: The calculated duration of the Eye event in sec.msec-usec
        #: format.
        self.duration = None

        #: The amplitude of the Saccade in the horizonatal direction.
        #: Usually specified in visual degrees.
        self.amplitude_x

        #: The amplitude of the Saccade in the vertical direction.
        #: Usually specified in visual degrees.
        self.amplitude_y

        #: The angle of the Saccade based on the start and end gaze positions.
        #: Usually specified in degrees.
        self.angle

        #: The calibrated horizontal eye position at the start of the eye event
        #: on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.start_gaze_x = None

        #: The calibrated vertical eye position at the start of the eye event on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.start_gaze_y = None

        #: The calculated point of gaze in depth at the start of the eye event. Generally this can only be
        #: provided if binocular reporting is being performed.
        self.start_gaze_z = None

        #: The horizontal angle of eye the relative to the head at the start of the eye event.
        self.start_angle_x = None

        #: The vertical angle of eye the relative to the head at the start of the eye event.
        self.start_angle_y = None

        #: The non-calibrated x position of the calculated eye 'center'
        #: on the camera sensor image at the start of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.start_raw_x = None

        #: The non-calibrated y position of the calculated eye 'center'
        #: on the camera sensor image at the start of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.start_raw_y = None

        #: A measure related to pupil size or diameter at the start of the eye event.
        #: The attribute pupil_measure1_type defines what type the measure represents.
        self.start_pupil_measure1 = None

        #: The type of pupil size or shape information provided in the pupil_measure1
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.start_pupil_measure1_type = None

        #: A second measure related to pupil size or diameter at the start of the eye event.
        #: The attribute pupil_measure2_type defines what type the measure represents.
        self.start_pupil_measure2 = None

        #: The type of pupil size or shape information provided in the pupil_measure2
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.start_pupil_measure2_type = None

        #: Horizontal pixels per visual degree for this eye position at the start of the eye event
        #: as reported by the eye tracker.
        self.start_ppd_x = None

        #: Vertical pixels per visual degree for this eye position at the start of the eye event
        #: as reported by the eye tracker.
        self.start_ppd_y = None

        #: Horizontal velocity of the eye at the start of the eye event;
        #: as reported by the eye tracker.
        self.start_velocity_x = None

        #: Vertical velocity of the eye at the start of the eye event;
        #: as reported by the eye tracker.
        self.start_velocity_y = None

        #: The calibrated horizontal eye position at the end of the eye event
        #: on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.end_gaze_x = None

        #: The calibrated vertical eye position at the end of the eye event on the calibration plane.
        #: This value is specified in Display Coordinate Type Units.
        self.end_gaze_y = None

        #: The calculated point of gaze in depth at the end of the eye event. Generally this can only be
        #: provided if binocular reporting is being performed.
        self.end_gaze_z = None

        #: The horizontal angle of eye the relative to the head at the end of the eye event.
        self.end_angle_x = None

        #: The vertical angle of eye the relative to the head at the end of the eye event.
        self.end_angle_y = None

        #: The non-calibrated x position of the calculated eye 'center'
        #: on the camera sensor image at the end of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.end_raw_x = None

        #: The non-calibrated y position of the calculated eye 'center'
        #: on the camera sensor image at the end of the eye event,
        #: factoring in any corneal reflection adjustments.
        #: This is typically reported in some arbitrary unit space that
        #: often has sub-pixel resolution due to image processing techniques
        #: being applied.
        self.end_raw_y = None

        #: A measure related to pupil size or diameter at the end of the eye event.
        #: The attribute pupil_measure1_type defines what type the measure represents.
        self.end_pupil_measure1 = None

        #: The type of pupil size or shape information provided in the pupil_measure1
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.end_pupil_measure1_type = None

        #: A second measure related to pupil size or diameter at the end of the eye event.
        #: The attribute pupil_measure2_type defines what type the measure represents.
        self.end_pupil_measure2 = None

        #: The type of pupil size or shape information provided in the pupil_measure2
        #: attribute. Several possible pupil_measure types available:
        #:
        #: * EyeTrackerConstants.PUPIL_AREA
        #: * EyeTrackerConstants.PUPIL_DIAMETER
        #: * EyeTrackerConstants.PUPIL_AREA_MM
        #: * EyeTrackerConstants.PUPIL_DIAMETER_MM
        #: * EyeTrackerConstants.PUPIL_WIDTH
        #: * EyeTrackerConstants.PUPIL_HEIGHT
        #: * EyeTrackerConstants.PUPIL_WIDTH_MM
        #: * EyeTrackerConstants.PUPIL_HEIGHT_MM
        #: * EyeTrackerConstants.PUPIL_MAJOR_AXIS
        #: * EyeTrackerConstants.PUPIL_MINOR_AXIS
        self.end_pupil_measure2_type = None

        #: Horizontal pixels per visual degree for this eye position at the end of the eye event
        #: as reported by the eye tracker.
        self.end_ppd_x = None

        #: Vertical pixels per visual degree for this eye position at the end of the eye event
        #: as reported by the eye tracker.
        self.end_ppd_y = None

        #: Horizontal velocity of the eye at the end of the eye event;
        #: as reported by the eye tracker.
        self.end_velocity_x = None

        #: Vertical velocity of the eye at the end of the eye event;
        #: as reported by the eye tracker.
        self.end_velocity_y = None

        #: 2D Velocity of the eye at the end of the eye event;
        #: as reported by the eye tracker.
        self.end_velocity_xy = None

        #: Average Horizontal velocity of the eye during the eye event;
        #: as reported by the eye tracker.
        self.average_velocity_x = None

        #: Average Vertical velocity of the eye during the eye event;
        #: as reported by the eye tracker.
        self.average_velocity_y = None

        #: Average 2D Velocity of the eye at the during the eye event;
        #: as reported by the eye tracker.
        self.average_velocity_xy = None

        #: Peak Horizontal velocity of the eye during the eye event;
        #: as reported by the eye tracker.
        self.peak_velocity_x = None

        #: Peak Vertical velocity of the eye during the eye event;
        #: as reported by the eye tracker.
        self.peak_velocity_y = None

        #: Peak 2D Velocity of the eye at the during the eye event;
        #: as reported by the eye tracker.
        self.peak_velocity_xy = None

        #: An available status byte for the eye tracker event.
        #: Meaning or use is completely tracker dependent.
        self.status = None

        DeviceEvent.__init__(self, *args, **kwargs)


################### Blink Event Types ##########################
#
class BlinkStartEvent(EyeTrackerEvent):
    _newDataTypes = [
        # The eye type that the fixation is from. Valid values are:
        ('eye', 'u1'),
        #   EyeTrackerConstants.LEFT_EYE
        #   EyeTrackerConstants.RIGHT_EYE
        #   EyeTrackerConstants.BINOCULAR_AVERAGED
        #   EyeTrackerConstants.SIMULATED_MONOCULAR


        # An available status byte for the eye tracker blink start event.
        ('status', 'u1')
        # Meaning is completely tracker dependent.
    ]
    __slots__ = [e[0] for e in _newDataTypes]

    EVENT_TYPE_ID = EventConstants.BLINK_START
    EVENT_TYPE_STRING = EventConstants.getName(EVENT_TYPE_ID)
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    def __init__(self, *args, **kwargs):

        #: The eye type that the event is from. Valid values are:
        #: EyeTrackerConstants.LEFT_EYE
        #: EyeTrackerConstants.RIGHT_EYE
        #: EyeTrackerConstants.MONOCULAR
        #: EyeTrackerConstants.SIMULATED_MONOCULAR
        self.eye = None

        #: An available status byte for the eye tracker event.
        #: Meaning or use is completely tracker dependent.
        self.status = None

        DeviceEvent.__init__(self, *args, **kwargs)


class BlinkEndEvent(EyeTrackerEvent):
    _newDataTypes = [
        # The eye type that the fixation is from. Valid values are:
        ('eye', 'u1'),
        #   EyeTrackerConstants.LEFT_EYE
        #   EyeTrackerConstants.RIGHT_EYE
        #   EyeTrackerConstants.BINOCULAR_AVERAGED
        #   EyeTrackerConstants.SIMULATED_MONOCULAR


        ('duration', 'f4'),  # The duration of the blink event.

        # An available status byte for the eye tracker blink start event.
        ('status', 'u1')
        # Meaning is completely tracker dependent.
    ]

    EVENT_TYPE_ID = EventConstants.BLINK_END
    EVENT_TYPE_STRING = EventConstants.getName(EVENT_TYPE_ID)
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):

        #: The eye type that the event is from. Valid values are:
        #: EyeTrackerConstants.LEFT_EYE
        #: EyeTrackerConstants.RIGHT_EYE
        #: EyeTrackerConstants.MONOCULAR
        #: EyeTrackerConstants.SIMULATED_MONOCULAR
        self.eye = None

        #: The calculated duration of the Eye event in sec.msec-usec
        #: format.
        self.duration = None

        #: An available status byte for the eye tracker event.
        #: Meaning or use is completely tracker dependent.
        self.status = None

        DeviceEvent.__init__(self, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""A Backend class defines the core low-level functions required by a Window
class, such as the ability to create an OpenGL context and flip the window.

Users simply call visual.Window(..., winType='pyglet') and the winType is then
used by backends.getBackend(winType) which will locate the appropriate class
and initialize an instance using the attributes of the Window.
"""

from __future__ import absolute_import, print_function
import sys
import os
import numpy as np

import psychopy
from psychopy import logging, event, platform_specific, constants
from psychopy.visual import window
from psychopy.tools.attributetools import attributeSetter
from .gamma import setGamma, setGammaRamp, getGammaRamp, getGammaRampSize
from .. import globalVars
from ._base import BaseBackend

import pyglet
# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+
pyglet.options['debug_gl'] = False
GL = pyglet.gl

retinaContext = None  # it will be set to an actual context if needed

# get the default display
if pyglet.version < '1.4':
    _default_display_ = pyglet.window.get_platform().get_default_display()
else:
    _default_display_ = pyglet.canvas.get_display()


class PygletBackend(BaseBackend):
    """The pyglet backend is the most used backend. It has no dependencies
    or C libs that need compiling, but may not be as fast or efficient as libs
    like GLFW.
    """
    GL = pyglet.gl
    winTypeName = 'pyglet'

    def __init__(self, win, backendConf=None):
        """Set up the backend window according the params of the PsychoPy win

        Parameters
        ----------
        win : `psychopy.visual.Window` instance
            PsychoPy Window (usually not fully created yet).
        backendConf : `dict` or `None`
            Backend configuration options. Options are specified as a dictionary
            where keys are option names and values are settings. For this
            backend the following options are available:

            * `bpc` (`array_like` of `int`) Bits per color (R, G, B).
            * `depthBits` (`int`) Framebuffer (back buffer) depth bits.
            * `stencilBits` (`int`) Framebuffer (back buffer) stencil bits.

        Examples
        --------
        Create a window using the Pyglet backend and specify custom options::

            import psychopy.visual as visual

            options = {'bpc': (8, 8, 8), 'depthBits': 24, 'stencilBits': 8}
            win = visual.Window(winType='pyglet', backendOptions=options)

        """
        BaseBackend.__init__(self, win)  # sets up self.win=win as weakref

        # if `None`, change to `dict` to extract options
        backendConf = backendConf if backendConf is not None else {}

        if not isinstance(backendConf, dict):  # type check on options
            raise TypeError(
                'Object passed to `backendConf` must be type `dict`.')

        self._TravisTesting = (os.environ.get('TRAVIS') == 'true')

        self._gammaErrorPolicy = win.gammaErrorPolicy
        self._origGammaRamp = None
        self._rampSize = None

        vsync = 0

        # provide warning if stereo buffers are requested but unavailable
        if win.stereo and not GL.gl_info.have_extension('GL_STEREO'):
            logging.warning(
                'A stereo window was requested but the graphics '
                'card does not appear to support GL_STEREO')
            win.stereo = False

        if sys.platform=='darwin' and not win.useRetina and pyglet.version >= "1.3":
            raise ValueError("As of PsychoPy 1.85.3 OSX windows should all be "
                             "set to useRetina=True (or remove the argument). "
                             "Pyglet 1.3 appears to be forcing "
                             "us to use retina on any retina-capable screen "
                             "so setting to False has no effect.")

        # window framebuffer configuration
        bpc = backendConf.get('bpc', (8, 8, 8))
        if isinstance(bpc, int):
            win.bpc = (bpc, bpc, bpc)
        else:
            win.bpc = bpc

        win.depthBits = int(backendConf.get('depthBits', 8))

        if win.allowStencil:
            win.stencilBits = int(backendConf.get('stencilBits', 8))
        else:
            win.stencilBits = 0

        # multisampling
        sample_buffers = 0
        aa_samples = 0

        if win.multiSample:
            sample_buffers = 1
            # get maximum number of samples the driver supports
            max_samples = (GL.GLint)()
            GL.glGetIntegerv(GL.GL_MAX_SAMPLES, max_samples)

            if (win.numSamples >= 2) and (
                        win.numSamples <= max_samples.value):
                # NB - also check if divisible by two and integer?
                aa_samples = win.numSamples
            else:
                logging.warning(
                    'Invalid number of MSAA samples provided, must be '
                    'integer greater than two. Disabling.')
                win.multiSample = False

        if pyglet.version < '1.4':
            allScrs = _default_display_.get_screens()
        else:
            allScrs = _default_display_.get_screens()

        # Screen (from Exp Settings) is 1-indexed,
        # so the second screen is Screen 1
        if len(allScrs) < int(win.screen) + 1:
            logging.warn("Requested an unavailable screen number - "
                         "using first available.")
            thisScreen = allScrs[0]
        else:
            thisScreen = allScrs[win.screen]
            if win.autoLog:
                logging.info('configured pyglet screen %i' % win.screen)

        # options that the user might want
        config = GL.Config(depth_size=win.depthBits,
                           double_buffer=True,
                           sample_buffers=sample_buffers,
                           samples=aa_samples,
                           stencil_size=win.stencilBits,
                           stereo=win.stereo,
                           vsync=vsync,
                           red_size=win.bpc[0],
                           green_size=win.bpc[1],
                           blue_size=win.bpc[2])

        # check if we can have this configuration
        validConfigs = thisScreen.get_matching_configs(config)
        if not validConfigs:
            # check which configs are invalid for the display
            raise RuntimeError(
                "Specified window configuration is not supported by this "
                "display.")

        # if fullscreen check screen size
        if win._isFullScr:
            win._checkMatchingSizes(win.clientSize, [thisScreen.width,
                                                  thisScreen.height])
            w = h = None
        else:
            w, h = win.clientSize
        if win.allowGUI:
            style = None
        else:
            style = 'borderless'
        try:
            self.winHandle = pyglet.window.Window(
                    width=w, height=h,
                    caption="PsychoPy",
                    fullscreen=win._isFullScr,
                    config=config,
                    screen=thisScreen,
                    style=style)
        except pyglet.gl.ContextException:
            # turn off the shadow window an try again
            pyglet.options['shadow_window'] = False
            self.winHandle = pyglet.window.Window(
                    width=w, height=h,
                    caption="PsychoPy",
                    fullscreen=self._isFullScr,
                    config=config,
                    screen=thisScreen,
                    style=style)
            logging.warning(
                "Pyglet shadow_window has been turned off. This is "
                "only an issue for you if you need multiple "
                "stimulus windows, in which case update your "
                "graphics card and/or graphics drivers.")

        if sys.platform == 'win32':
            # pyHook window hwnd maps to:
            # pyglet 1.14 -> window._hwnd
            # pyglet 1.2a -> window._view_hwnd
            if pyglet.version > "1.2":
                win._hw_handle = self.winHandle._view_hwnd
            else:
                win._hw_handle = self.winHandle._hwnd

            self._frameBufferSize = win.clientSize
        elif sys.platform == 'darwin':
            if win.useRetina:
                global retinaContext
                retinaContext = self.winHandle.context._nscontext
                view = retinaContext.view()
                bounds = view.convertRectToBacking_(view.bounds()).size
                if win.clientSize[0] == bounds.width:
                    win.useRetina = False  # the screen is not a retina display
                self._frameBufferSize = np.array([int(bounds.width), int(bounds.height)])
            else:
                self._frameBufferSize = win.clientSize
            try:
                # python 32bit (1.4. or 1.2 pyglet)
                win._hw_handle = self.winHandle._window.value
            except Exception:
                # pyglet 1.2 with 64bit python?
                win._hw_handle = self.winHandle._nswindow.windowNumber()
        elif sys.platform.startswith('linux'):
            win._hw_handle = self.winHandle._window
            self._frameBufferSize = win.clientSize

        if win.useFBO:  # check for necessary extensions
            if not GL.gl_info.have_extension('GL_EXT_framebuffer_object'):
                msg = ("Trying to use a framebuffer object but "
                       "GL_EXT_framebuffer_object is not supported. Disabled")
                logging.warn(msg)
                win.useFBO = False
            if not GL.gl_info.have_extension('GL_ARB_texture_float'):
                msg = ("Trying to use a framebuffer object but "
                       "GL_ARB_texture_float is not supported. Disabling")
                logging.warn(msg)
                win.useFBO = False

        if pyglet.version < "1.2" and sys.platform == 'darwin':
            platform_specific.syncSwapBuffers(1)

        # add these methods to the pyglet window
        self.winHandle.setGamma = setGamma
        self.winHandle.setGammaRamp = setGammaRamp
        self.winHandle.getGammaRamp = getGammaRamp
        self.winHandle.set_vsync(True)
        self.winHandle.on_text = self.onText
        self.winHandle.on_text_motion = self.onCursorKey
        self.winHandle.on_key_press = self.onKey
        self.winHandle.on_mouse_press = event._onPygletMousePress
        self.winHandle.on_mouse_release = event._onPygletMouseRelease
        self.winHandle.on_mouse_scroll = event._onPygletMouseWheel
        if not win.allowGUI:
            # make mouse invisible. Could go further and make it 'exclusive'
            # (but need to alter x,y handling then)
            self.winHandle.set_mouse_visible(False)
        self.winHandle.on_resize = _onResize  # avoid circular reference
        if not win.pos:
            # work out where the centre should be 
            if win.useRetina:
                win.pos = [(thisScreen.width - win.clientSize[0]/2) / 2,
                            (thisScreen.height - win.clientSize[1]/2) / 2]
            else:
                win.pos = [(thisScreen.width - win.clientSize[0]) / 2,
                            (thisScreen.height - win.clientSize[1]) / 2]
        if not win._isFullScr:
            # add the necessary amount for second screen
            self.winHandle.set_location(int(win.pos[0] + thisScreen.x),
                                        int(win.pos[1] + thisScreen.y))

        try:  # to load an icon for the window
            iconFile = os.path.join(psychopy.prefs.paths['resources'],
                                    'psychopy.ico')
            icon = pyglet.image.load(filename=iconFile)
            self.winHandle.set_icon(icon)
        except Exception:
            pass  # doesn't matter

        # store properties of the system
        self._driver = pyglet.gl.gl_info.get_renderer()

    @property
    def frameBufferSize(self):
        """Size of the presently active framebuffer in pixels (w, h)."""
        return self._frameBufferSize

    @property
    def shadersSupported(self):
        # on pyglet shaders are fine so just check GL>2.0
        return pyglet.gl.gl_info.get_version() >= '2.0'

    def swapBuffers(self, flipThisFrame=True):
        """Performs various hardware events around the window flip and then
        performs the actual flip itself (assuming that flipThisFrame is true)

        :param flipThisFrame: setting this to False treats this as a frame but
            doesn't actually trigger the flip itself (e.g. because the device
            needs multiple rendered frames per flip)
        """
        # make sure this is current context
        if globalVars.currWindow != self:
            self.winHandle.switch_to()
            globalVars.currWindow = self

        GL.glTranslatef(0.0, 0.0, -5.0)

        for dispatcher in self.win._eventDispatchers:
            try:
                dispatcher.dispatch_events()
            except:
                dispatcher._dispatch_events()

        # this might need to be done even more often than once per frame?
        self.winHandle.dispatch_events()

        # for pyglet 1.1.4 you needed to call media.dispatch for
        # movie updating
        if pyglet.version < '1.2':
            pyglet.media.dispatch_events()  # for sounds to be processed
        if flipThisFrame:
            self.winHandle.flip()

    def setMouseVisibility(self, visibility):
        self.winHandle.set_mouse_visible(visibility)

    def setCurrent(self):
        """Sets this window to be the current rendering target.

        Returns
        -------
        bool
            ``True`` if the context was switched from another. ``False`` is
            returned if ``setCurrent`` was called on an already current window.

        """
        if self != globalVars.currWindow:
            self.winHandle.switch_to()
            globalVars.currWindow = self

            return True

        return False

    def dispatchEvents(self):
        """Dispatch events to the event handler (typically called on each frame)

        :return:
        """
        wins = _default_display_.get_windows()
        for win in wins:
            win.dispatch_events()

    def onKey(self, evt, modifiers):
        "Check for tab key then pass all events to event package"
        if evt is not None:
            thisKey = pyglet.window.key.symbol_string(evt).lower()
            if thisKey == 'tab':
                self.onText('\t')
            event._onPygletKey(evt, modifiers)

    def onText(self, evt):
        """Retrieve the character event(s?) for this window"""
        if evt is not None:
            currentEditable = self.win.currentEditable
            if currentEditable:
                currentEditable._onText(evt)
            event._onPygletText(evt)  # duplicate the event to the psychopy.events lib

    def onCursorKey(self, evt):
        """Processes the events from pyglet.window.on_text_motion

        which is keys like cursor, delete, backspace etc."""
        currentEditable = self.win.currentEditable
        if currentEditable:
            keyName = pyglet.window.key.motion_string(evt)
            currentEditable._onCursorKeys(keyName)

    def onResize(self, width, height):
        _onResize(width, height)

    @attributeSetter
    def gamma(self, gamma):
        self.__dict__['gamma'] = gamma
        if self._TravisTesting:
            return
        if self._origGammaRamp is None:  # get the original if we haven't yet
            self._getOrigGammaRamp()
        if gamma is not None:
            setGamma(
                screenID=self.screenID,
                newGamma=gamma,
                rampSize=self._rampSize,
                driver=self._driver,
                xDisplay=self.xDisplay,
                gammaErrorPolicy=self._gammaErrorPolicy
            )

    @attributeSetter
    def gammaRamp(self, gammaRamp):
        """Gets the gamma ramp or sets it to a new value (an Nx3 or Nx1 array)
        """
        self.__dict__['gammaRamp'] = gammaRamp
        if self._TravisTesting:
            return
        if self._origGammaRamp is None:  # get the original if we haven't yet
            self._getOrigGammaRamp()
        setGammaRamp(
            self.screenID,
            gammaRamp,
            nAttempts=3,
            xDisplay=self.xDisplay,
            gammaErrorPolicy=self._gammaErrorPolicy
        )

    def getGammaRamp(self):
        return getGammaRamp(self.screenID, self.xDisplay,
                            gammaErrorPolicy=self._gammaErrorPolicy)

    def getGammaRampSize(self):
        return getGammaRampSize(self.screenID, self.xDisplay,
                                gammaErrorPolicy=self._gammaErrorPolicy)

    def _getOrigGammaRamp(self):
        """This is just used to get origGammaRamp and will populate that if
        needed on the first call"""
        if self._origGammaRamp is None:
            self._origGammaRamp = self.getGammaRamp()
            self._rampSize = self.getGammaRampSize()
        else:
            return self._origGammaRamp

    @property
    def screenID(self):
        """Returns the screen ID or device context (depending on the platform)
        for the current Window
        """
        if sys.platform == 'win32':
            scrBytes = self.winHandle._dc
            if constants.PY3:
                try:
                    _screenID = 0xFFFFFFFF & int.from_bytes(scrBytes, byteorder='little')
                except TypeError:
                    _screenID = 0xFFFFFFFF & scrBytes
            else:
                try:
                    _screenID = 0xFFFFFFFF & scrBytes
                except TypeError:
                    _screenID = scrBytes

        elif sys.platform == 'darwin':
            try:
                _screenID = self.winHandle._screen.id  # pyglet1.2alpha1
            except AttributeError:
                _screenID = self.winHandle._screen._cg_display_id  # pyglet1.2
        elif sys.platform.startswith('linux'):
            _screenID = self.winHandle._x_screen_id
        return _screenID

    @property
    def xDisplay(self):
        """On X11 systems this returns the XDisplay being used and None on all
        other platforms"""
        if sys.platform.startswith('linux'):
            return self.winHandle._x_display

    def close(self):
        """Close the window and uninitialize the resources
        """
        # Check if window has device context and is thus not closed
        if self.winHandle.context is None:
            return

        # restore the gamma ramp that was active when window was opened
        if self._origGammaRamp is not None:
            self.gammaRamp = self._origGammaRamp

        _hw_handle = None
        try:
            _hw_handle = self.win._hw_handle
            self.winHandle.close()
        except Exception:
            pass
        # If iohub is running, inform it to stop looking for this win id
        # when filtering kb and mouse events (if the filter is enabled of
        # course)
        try:
            if window.IOHUB_ACTIVE and _hw_handle:
                from psychopy.iohub.client import ioHubConnection
                conn = ioHubConnection.ACTIVE_CONNECTION
                conn.unregisterWindowHandles(_hw_handle)
        except Exception:
            pass

    def setFullScr(self, value):
        """Sets the window to/from full-screen mode"""
        self.winHandle.set_fullscreen(value)


def _onResize(width, height):
    """A default resize event handler.

    This default handler updates the GL viewport to cover the entire
    window and sets the ``GL_PROJECTION`` matrix to be orthogonal in
    window space.  The bottom-left corner is (0, 0) and the top-right
    corner is the width and height of the :class:`~psychopy.visual.Window`
    in pixels.

    Override this event handler with your own to create another
    projection, for example in perspective.
    """
    global retinaContext

    if height == 0:
        height = 1

    if retinaContext is not None:
        view = retinaContext.view()
        bounds = view.convertRectToBacking_(view.bounds()).size
        back_width, back_height = (int(bounds.width), int(bounds.height))
    else:
        back_width, back_height = width, height

    GL.glViewport(0, 0, back_width, back_height)
    GL.glMatrixMode(GL.GL_PROJECTION)
    GL.glLoadIdentity()
    GL.glOrtho(-1, 1, -1, 1, -1, 1)
    # GL.gluPerspective(90, 1.0 * width / height, 0.1, 100.0)
    GL.glMatrixMode(GL.GL_MODELVIEW)
    GL.glLoadIdentity()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function

# from future import standard_library
# standard_library.install_aliases()
from builtins import str
from builtins import object
import codecs
import pickle
import psychopy.data

######### Begin Compatibility Class Definitions #########


class _oldStyleBaseTrialHandler(object):
    """Please excuse these ugly kluges, but in order to unpickle
        psydat pickled trial handlers that were created using the old-style
        (pre python 2.2) class, original classes have to be defined.
    """
    pass


class _oldStyleBaseStairHandler(object):
    """Stubbed compapatibility class for StairHandler"""
    pass


class _oldStyleTrialHandler(object):
    """Stubbed compapatibility class for TrialHandler"""
    pass


class _oldStyleStairHandler(object):
    """Stubbed compapatibility class for StairHandler"""
    pass


class _oldStyleMultiStairHandler(object):
    """Stubbed compapatibility class for MultiStairHandler"""
    pass
######### End Compatibility Class Definitions #########


def _convertToNewStyle(newClass, oldInstance):
    """Converts un-pickled old-style compatibility classes to new-style ones
       by initializing a new-style class and copying the old compatibility
       instance's attributes.
    """
    # if the oldInstance was an ExperimentHandler it wouldn't throw an error
    # related to itself, but to the underlying loops within it. So check if we
    # have that and then do imports on each loop
    if oldInstance.__class__.__name__ == 'ExperimentHandler':
        newHandler = psychopy.data.ExperimentHandler()
        # newClass()  # Init a new new-style object
    else:
        newHandler = newClass([], 0)  # Init a new new-style object
    for thisAttrib in dir(oldInstance):
        # can handle each attribute differently
        if 'instancemethod' in str(type(getattr(oldInstance, thisAttrib))):
            # this is a method
            continue
        elif thisAttrib == '__weakref__':
            continue
        else:
            value = getattr(oldInstance, thisAttrib)
            setattr(newHandler, thisAttrib, value)
    return newHandler


def fromFile(filename):
    """In order to switch experiment handler to the new-style
    (post-python 2.2, circa 2001) classes, this is a proof-of-concept loader
    based on tools.filetools.fromFile that will load psydat files created
    with either new or old style TrialHandlers or StairHandlers.

    Since this is really just an example, it probably [hopefully] won't be
    incorporated into upstream code, but it will work.

    The method will try to load the file using a normal new-style Pickle
    loader; however, if a type error occurs it will temporarily replace the
    new-style class with a stubbed version of the old-style class and will
    then instantiate a fresh new-style class with the original attributes.
    """
    with codecs.open(filename, 'rb') as f:
        try:
            # Try to load the psydat file into the new-style class.
            contents = pickle.load(f)
        except UnicodeDecodeError:
            f.seek(0)  # reset to start of file to try again
            contents = pickle.load(f, encoding='latin1')  # python 2 data files
        except TypeError as e:
            f.seek(0)  # reset to start of file to try again
            # check er as string for one of our handlers
            errStr = "{}".format(e)
            name = ''
            for thisHandler in ['TrialHandler','StairHandler','MultiStairHandler']:
                if thisHandler in errStr:
                    name = thisHandler
            # if error is from something else try to deduce the class
            if not name:
                name = e.args[1].__name__
            # then process accordingly
            if name == 'TrialHandler':
                currentHandler = psychopy.data.TrialHandler
                # Temporarily replace new-style class
                psychopy.data.TrialHandler = _oldStyleTrialHandler
                oldContents = pickle.load(f)
                psychopy.data.TrialHandler = currentHandler
                contents = _convertToNewStyle(psychopy.data.TrialHandler,
                                              oldContents)
            elif name == 'StairHandler':
                currentHandler = psychopy.data.StairHandler
                # Temporarily replace new-style class
                psychopy.data.StairHandler = _oldStyleStairHandler
                oldContents = pickle.load(f)
                psychopy.data.StairHandler = currentHandler
                contents = _convertToNewStyle(
                    psychopy.data.StairHandler, oldContents)
            elif name == '':
                newStair = psychopy.data.StairHandler
                newMulti = psychopy.data.MultiStairHandler
                # Temporarily replace new-style class
                psychopy.data.StairHandler = _oldStyleStairHandler
                # Temporarily replace new-style class
                psychopy.data.MultiStairHandler = _oldStyleMultiStairHandler
                oldContents = pickle.load(f)
                psychopy.data.MultiStairHandler = newMulti
                # Temporarily replace new-style class:
                psychopy.data.StairHandler = newStair
                contents = _convertToNewStyle(
                    psychopy.data.MultiStairHandler, oldContents)
            else:
                raise TypeError("Didn't recognize %s" % name)
    return contents


def checkCompatibility(old, new, prefs=None, fix=True):
    """Check for known compatibility issue between a pair of versions and fix
    automatically if possible. (This facility, and storing the most recently
    run version of the app, was added in version 1.74.00)

    usage::

        ok, msg =  checkCompatibility(old, new, prefs, fix=True)

    prefs is a standard psychopy preferences object. It isn't needed by all
    checks but may be useful.
    This function can be used simply to check for issues by setting fix=False
    """
    if old == new:
        return 1, ""  # no action needed
    if old > new:  # switch them over
        old, new = new, old

    msg = "From %s to %s:" % (old, new)
    warning = False
    if old[0:4] < '1.74':
        msg += ("\n\nThere were many changes in version 1.74.00 that will "
                "break\ncompatibility with older versions. Make sure you read"
                " the changelog carefully\nbefore using this version. Do not "
                "upgrade to this version halfway through an experiment.\n")
        if fix and 'PatchComponent' not in prefs.builder['hiddenComponents']:
            prefs.builder['hiddenComponents'].append('PatchComponent')
        warning = True
    if not warning:
        msg += "\nNo known compatibility issues"

    return (not warning), msg



from psychopy import visual
from psychopy.hardware import keyboard

# Make window
win = visual.Window(units="height")

# Make instructions textbox
instr = visual.TextBox2(win,
    text="Press any key...",
    font="Open Sans", letterHeight=0.1,
    pos=(0, 0.2))
# Make key name textbox
output = visual.TextBox2(win,
    text="No key pressed yet",
    font="Open Sans", letterHeight=0.1,
    pos=(0, -0.2))
# Make keyboard object
kb = keyboard.Keyboard()
# Listen for keypresses until escape is pressed
keys = kb.getKeys()
while 'escape' not in keys:
    # Draw stimuli
    instr.draw()
    output.draw()
    win.flip()
    # Check keypresses
    keys = kb.getKeys()
    if keys:
        # If a key was pressed, display it
        output.text = f"That key was `{keys[-1].name}`. Press another key."
# End the demo
win.close()



from __future__ import print_function
from builtins import object

from os import path
import shutil
import py_compile
from tempfile import mkdtemp
import codecs
import pytest
import locale
import time

import psychopy.experiment
from psychopy import prefs
from psychopy.app import psychopyApp
from psychopy.app.builder.dialogs import DlgComponentProperties
from psychopy.app.builder.validators import CodeSnippetValidator
from psychopy.experiment import Param

# Jeremy Gray March 2011

# caveats when comparing files:
# - dicts have no defined order, can load and save differently: use a
#   known-diff file to suppress boring errors.  This situation was
#   addressed in 7e2c72a for stimOut by sorting the keys
# - namespace.makeValid() can change var names from the orig demos,
#   but should not do so from a load-save-load because only the first
#   load should change things

allComponents = psychopy.experiment.getComponents(fetchIcons=False)
import wx

class Test_BuilderFrame(object):
    """This test fetches all standard components and checks that, with default
    settings, they can be added to a Routine and result in a script that compiles
    """

    def setup(self):
        self.app = psychopyApp._app

        self.builder = self.app.newBuilderFrame()
        self.exp = self.builder.exp
        self.here = path.abspath(path.dirname(__file__))
        self.tmp_dir = mkdtemp(prefix='psychopy-tests-app')
        self.exp.addRoutine('testRoutine')
        self.testRoutine = self.exp.routines['testRoutine']
        self.exp.flow.addRoutine(self.testRoutine, 0)

    def teardown(self):
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

    def test_BuilderFrame(self):
        """Tests of the Builder frame. We can call dialog boxes using
        a timeout (will simulate OK being pressed)
        """
        builderView = self.app.newBuilderFrame()

        expfile = path.join(prefs.paths['tests'],
                            'data', 'test001EntryImporting.psyexp')
        builderView.fileOpen(filename=expfile)
        builderView.setExperimentSettings(timeout=500)
        builderView.isModified = False
        builderView.runFile()
        builderView.closeFrame()

    def _checkCompileWith(self, thisComp):
        """Adds the component to the current Routine and makes sure it still
        compiles
        """
        filename = thisComp.params['name'].val+'.py'
        filepath = path.join(self.tmp_dir, filename)

        self.testRoutine.addComponent(thisComp)
        #make sure the mouse code compiles

        # generate a script, similar to 'lastrun.py':
        buff = self.exp.writeScript() # is a StringIO object
        script = buff.getvalue()
        assert len(script) > 1500 # default empty script is ~2200 chars

        # save the script:
        f = codecs.open(filepath, 'w', 'utf-8')
        f.write(script)
        f.close()

        # compile the temp file to .pyc, catching error msgs (including no file at all):
        py_compile.compile(filepath, doraise=True)
        return filepath + 'c'

    def test_MessageDialog(self):
        """Test the message dialog
        """
        from psychopy.app.dialogs import MessageDialog
        dlg = MessageDialog(message="Just a test", timeout=500)
        ok = dlg.ShowModal()
        assert ok == wx.ID_OK

    def test_ComponentDialogs(self):
        """Test the message dialog
        """
        builderView = self.app.newBuilderFrame()
        componsPanel = builderView.componentButtons
        for thisComponName in list(componsPanel.components):
            # simulate clicking the button for each component
            assert componsPanel.onClick(thisComponName, timeout=500)
        builderView.isModified = False
        builderView.closeFrame()
        del builderView, componsPanel

    def test_param_validator(self):
        """Test the code validator for component parameters"""

        # Define 'tykes' - combinations of values likely to cause an error if certain features aren't working
        tykes = [
            {'fieldName': "brokenCode", 'param': Param(val="for + :", valType="code"), 'msg': "Python syntax error in field `{fieldName}`:  {param.val}"}, # Make sure it's picking up clearly broken code
            {'fieldName': "correctAns", 'param': Param(val="'space'", valType="code"), 'msg': ""}, # Single-element lists should not cause warning
        ]
        for tyke in tykes:
            # For each tyke, create a dummy environment
            parent = DlgComponentProperties(
                frame=self.builder, title='Param Testing',
                params={tyke['fieldName']: tyke['param']}, order=[],
                testing=True)
            # Set validator and validate
            parent.SetValidator(CodeSnippetValidator(tyke['fieldName']))
            parent.Validate()
            # Does the message delivered by the validator match what is expected?
            warnings = [w for w in list(parent.warningsDict.values()) if w] or ['']
            msg = warnings[0]
            assert msg == tyke['msg'].format(**tyke)
            # Cleanup
            parent.Destroy()